/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.chemistry;

import com.ggasoftware.indigo.Indigo;
import com.ggasoftware.indigo.IndigoObject;
import com.google.common.base.Preconditions;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalStructure;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

/**
 * <p> Factory for creating new entities of Type
 * <code>ChemicalStructure</code>.
 *
 * <p>Prefer this over
 * <code>new ChemicalStructure()</code> as it also generates the structure key
 * (InChi Key) automatically based on the given
 * <code>structureData</code>.</p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Component("chemicalStructureFactory")
public class ChemicalStructureFactory {

    private static final XLogger logger = XLoggerFactory.getXLogger("ChemicalStructureFactory");
    private ChemicalFormatConverter formatConverter;
    @Autowired
    private ChemistryFormat chemistryFormat;

    public ChemicalStructureFactory(ChemistryFormat chemistryFormat) {
        this.chemistryFormat = chemistryFormat;
        this.formatConverter = new ChemicalFormatConverter();
    }

    protected ChemicalStructureFactory() {
        this.formatConverter = new ChemicalFormatConverter();
    }

    /**
     * <p> Creates a new
     * <code>ChemicalStructure</code> with
     * <code>structureData</code> and
     * <code>structureKey</code> properly set. </p>
     *
     * Uses Bingo Cartridge or CDK for generating InChikey.
     *
     * @param structureData SMILES or Molfile depending on your configuration
     * @return
     */
    public ChemicalStructure createChemicalStructure(String structureData) {

        logger.entry(structureData);
        Preconditions.checkNotNull(structureData);
        switch (chemistryFormat) {
            case SMI:
                structureData = formatConverter.toSmiles(structureData);
                break;
            case MOL:
                structureData = formatConverter.toMol(structureData);
                break;
            default:
                throw new IllegalArgumentException("ChemistryFormat not supported.");
        }
        ChemicalStructure structure;
        String structureKey = formatConverter.toStandardInchikey(structureData);
        logger.debug("Generated InChikey {} for {}.", structureKey, structureData);
        Indigo indigo = new Indigo();
        IndigoObject mol = indigo.loadMolecule(structureData);
        structure = new ChemicalStructure(structureKey, structureData, mol.molecularWeight());
        logger.debug("Succesfully generated ChemicalStructure instance for {} with structureKey {}.", structureData, structureKey);
        logger.exit(structure);
        return structure;
    }

    public ChemicalStructure createChemicalStructure(IndigoObject mol) {

        logger.entry(mol);
        Preconditions.checkNotNull(mol);
        String structureData;
        switch (chemistryFormat) {
            case SMI:
                mol.aromatize();
                structureData = mol.canonicalSmiles();
                break;
            case MOL:
                structureData = mol.molfile();
                break;
            default:
                throw new IllegalArgumentException("ChemistryFormat not supported.");
        }
        ChemicalStructure structure;
        // use molfile here as SMILES cause more erros to occure
        String structureKey = formatConverter.toStandardInchikey(mol.molfile());
        logger.debug("Generated InChikey {} for {}.", structureKey, structureData);
        structure = new ChemicalStructure(structureKey, structureData, mol.molecularWeight());
        logger.debug("Succesfully generated ChemicalStructure instance for {} with structureKey {}.", structureData, structureKey);
        logger.exit(structure);
        return structure;
    }
}
