/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.entity;

import java.util.Date;
import javax.persistence.Column;
import javax.persistence.MappedSuperclass;
import javax.persistence.PrePersist;
import javax.persistence.PreUpdate;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;
import javax.persistence.Version;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;
import org.springframework.security.core.context.SecurityContextHolder;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@MappedSuperclass
public abstract class BaseEntity {

    /*
     * Note that the setter is public by intention. This allows to bind the
     * version to a form (like a jsp page in spring mvc) and the entity generated
     * when submitting (POST) the form has the proper version set. else
     * optimistic locking will not be possible.
     */
    @Getter
    @Setter 
    @Version
    private Long version;

    @Getter
    @Setter(AccessLevel.PROTECTED)
    @Temporal(TemporalType.TIMESTAMP)
    @Column(name = "created")
    private Date created;

    @Getter
    @Setter(AccessLevel.PROTECTED)
    @Temporal(TemporalType.TIMESTAMP)
    @Column(name = "last_modified")
    private Date lastModified;

    @Getter
    @Setter(AccessLevel.PROTECTED)
    @Column(name = "created_by")
    private String createdBy;

    @Getter
    @Setter(AccessLevel.PROTECTED)
    @Column(name = "last_modified_by")
    private String lastModifiedBy;

    /*
     * Get unique id of this Entity
     */
    public abstract Long getId();

    /**
     * <p>
     * Automatically sets lastModified and lastModifiedBy before updating
     * the entity in the database
     * </p>
     * <p>
     * lastModifiedBy is only updated in case you use Spring security
     * </p>
     *
     */
    @PreUpdate
    public final void updateMetadata() {
        lastModified = new Date();
        if (SecurityContextHolder.getContext().getAuthentication() != null) {
            lastModifiedBy = SecurityContextHolder.getContext()
                    .getAuthentication().getName();
        }
    }

    /**
     * <p>
     * Automatically sets created and createdNy and
     * lastModified and lastModifiedBy before persisting the entity for the
     * first time.
     * </p>
     * <p>
     * createdBy By and lastModifiedBy is only set in case you use Spring security.
     * </p>
     *
     */
    @PrePersist
    public final void createMetadata() {
        created = new Date();
        lastModified = created;
        if (SecurityContextHolder.getContext().getAuthentication() != null) {
            final String userName = SecurityContextHolder.getContext()
                    .getAuthentication().getName();
            lastModifiedBy = userName;
            createdBy = userName;
        }
    }
}
