/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.entity;

import com.fasterxml.jackson.annotation.JsonIdentityInfo;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.ObjectIdGenerators;
import com.ggasoftware.indigo.Indigo;
import com.ggasoftware.indigo.IndigoObject;
import com.google.common.base.Preconditions;
import java.util.HashSet;
import java.util.Set;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.OrderBy;
import javax.persistence.SequenceGenerator;
import javax.persistence.Table;
import javax.persistence.Transient;
import lombok.Data;
import lombok.EqualsAndHashCode;

/**
 * <p>
 * A chemical compound consisting of one or more ChemicalStructures.
 * </p>
 * <p>
 * Each ChemicalStructures appears in the compound with a certain percentage.
 * </p>
 * <p>
 * A chemical compound like the chemical structure is considered to be more like
 * a Type. It can appear in other entities but it does not exist by itself and
 * each compound should only be in the database once (CAS is unique).
 * </p>
 * <p>
 * Note: You should use addCompostion to add a new composition instead of
 * chemicalCompound.getCompositions().add. In later case no check on total
 * percentage can be made.
 * </p>
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Entity
@Inheritance(strategy = InheritanceType.TABLE_PER_CLASS)
@Table(name="chemical_compound")
@JsonIdentityInfo(generator=ObjectIdGenerators.IntSequenceGenerator.class, property="@id")
@Data
@EqualsAndHashCode(callSuper = false, of = {"compoundName", "cas"})
public abstract class ChemicalCompound<T extends Containable> extends BaseEntity {

    private final static String PERCENTAGE_OVERFLOW_MESSAGE =
            "Total Percentage can't be higher than 100.";

    @Id
    @GeneratedValue(strategy = GenerationType.AUTO, generator = "chem_compound_seq_gen")
    @SequenceGenerator(name = "chem_compound_seq_gen",
    sequenceName = "seq_chem_compound", allocationSize = 1000)
    @Column(name = "chem_compound_id")
    private Long id;

    @Column(name = "compound_name", nullable = false)
    private String compoundName;

    @Column(name = "cas", length = 13)
    private String cas;

    // When set to FetchType.EAGER, if you delete this entity and then try to load
    // it again you get a EntityNotFoundException
    // http://stackoverflow.com/questions/11058085/entitynotfoundexception-deleted-entity-passed-to-persist-on-entitymanager-remov
    //@QueryInit({"chemicalCompound"})
    @OneToMany(fetch = FetchType.LAZY, mappedBy = "pk.compound",
        cascade = CascadeType.ALL, orphanRemoval = true)
    @OrderBy("pk.chemicalStructure.id ASC")
    //@Cache(usage = CacheConcurrencyStrategy.TRANSACTIONAL)
    private Set<ChemicalCompoundComposition> compositions = new HashSet<>();

    @Column(name = "compound_description")
    private String compoundDescription;

    @OneToMany(fetch = FetchType.LAZY, mappedBy="chemicalCompound",
        targetEntity=Containable.class, cascade = {CascadeType.PERSIST,
            CascadeType.REFRESH},
        orphanRemoval = true)
    //@Cache(usage = CacheConcurrencyStrategy.TRANSACTIONAL)
    private Set<T> containables = new HashSet<>();

    @JsonIgnore
    @OneToOne(fetch = FetchType.LAZY)
    private Role readRole;

    @Transient
    double totalPercentage = 0;

    public ChemicalCompound() {
    }

    /**
     *
     * @param id
     * @param compoundName
     * @param cas
     * @param description
     * @param composition
     * @throws IllegalArgumentException
     */
    public ChemicalCompound(Long id, String compoundName,
            String cas, String description,
            Set<ChemicalCompoundComposition> composition,
            Set<T> containables) {

        this.id = id;
        this.compoundName = compoundName;
        this.cas = cas;
        this.compoundDescription = description;
        this.containables = containables == null? new HashSet<T>() : containables;
        this.setCompositions(composition == null? new HashSet<ChemicalCompoundComposition>() : composition);
    }

    public ChemicalCompound(String compoundName, String cas,
            String description,
            Set<ChemicalCompoundComposition> composition,
            Set<T> containables) {
        this(null, compoundName, cas, description, composition, containables);
    }

    /**
     * <p>
     * Adds the given composition to this <code>ChemicalCompound</code>.
     * </p>
     * <p>
     * In case the total percentage of all compositions togeter is greater than
     * 100 an <code>IllegalArgumentException</code> is thrown.
     * </p>
     *
     * @param composition
     * @throws IllegalArgumentException
     */
    public void addComposition(ChemicalCompoundComposition composition) {
        if (composition.getPercentage() + getTotalPercentage() > 100) {
            throw new IllegalArgumentException(PERCENTAGE_OVERFLOW_MESSAGE);
        }
        compositions.add(composition);
    }

    public void addContainable(T containable){
        containables.add(containable);
    }

    /**
     * DON'T USE.
     *
     * @param readRole
     */
    public final void setReadRole(Role readRole){
        Preconditions.checkArgument(readRole.getRoleName().equals("read_" + this.getClass().getSimpleName()));
        this.readRole = readRole;
    }

    /**
     *
     * @param compositions
     * @throws IllegalArgumentException
     */
    public final void setCompositions(Set<ChemicalCompoundComposition> compositions) {
        double percentage = calcTotalPercentage(compositions);
        if (percentage > 100) {
            throw new IllegalArgumentException(PERCENTAGE_OVERFLOW_MESSAGE);
        }
        this.compositions = compositions;
        totalPercentage = percentage;
    }

    public void setTotalPercentage() {
        totalPercentage = calcTotalPercentage(compositions);
    }

    public double getTotalPercentage() {

        if (totalPercentage == 0) {
            setTotalPercentage();
        }
        return totalPercentage;
    }

    /**
     * <p>Generates a single molfile containing chemical structures from
     * all compositions as disconnected fragments.</p>
     *
     * @return the molfile of this compound
     */
    @JsonIgnore
    public String getMolfile() {
        Indigo indigo = new Indigo();
        IndigoObject mol = getIndigoMolecule(indigo);
        String molfile = mol.molfile();
        return molfile;
    }

    /**
     * <p>Generates a SMILES string containing chemical structures from
     * all compositions as separate components.</p>
     *
     * @return the molfile of this compound
     */
    @JsonIgnore
    public String getSmiles() {
        Indigo indigo = new Indigo();
        IndigoObject mol = getIndigoMolecule(indigo);
        String smiles = mol.canonicalSmiles();
        return smiles;
    }

    @JsonIgnore
    public IndigoObject getIndigoMolecule(Indigo indigo){

        IndigoObject mol = indigo.createMolecule();
        for (ChemicalCompoundComposition composition : compositions) {
            ChemicalStructure structure = composition.getChemicalStructure();
            String structureData = structure.getStructureData();
            mol.merge(indigo.loadMolecule(structureData));
            // sets the relative percentage of current composition using
            // inchikey as composition identifier
            mol.setProperty(structure.getStructureKey(), composition.getPercentage().toString());
        }
        mol.layout();
        return mol;
    }

    private double calcTotalPercentage(Set<ChemicalCompoundComposition> compositions) {
        double totalPercentage = 0;
        for (ChemicalCompoundComposition element : compositions) {
            if(element.getPercentage() != null) {
                totalPercentage += element.getPercentage();
            }
        }
        return totalPercentage;
    }
}
