/*
 *  Copyright (C) 2012 Joos Kiener
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.entity;

import com.fasterxml.jackson.annotation.JsonIdentityInfo;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.ObjectIdGenerators;
import java.util.ArrayList;
import java.util.List;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.OneToMany;
import javax.persistence.OrderBy;
import javax.persistence.SequenceGenerator;
import javax.persistence.Table;
import javax.persistence.Transient;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.Setter;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalFormatConverter;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemistryFormat;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.listener.StructureFormatListener;
import org.hibernate.annotations.Cache;


/**
 * <p>
 * Entity class for a Chemical Structure.
 * </p>
 * <p>
 * It contains the chemical structure data and a unique structure key which
 * identifies the chemical structure.
 * </p>
 * <p>
 * <b>This class must not be extended!</b> (entities can't be final hence this warning
 * in doc comments)
 * </p>
 * @author Joos Kiener
 */
@Entity
@EntityListeners(StructureFormatListener.class)
@Table(name = "chemical_structure")
@JsonIdentityInfo(generator=ObjectIdGenerators.IntSequenceGenerator.class, property="@id")
@EqualsAndHashCode(callSuper=false, of = {"structureKey"})
@Cache(usage = org.hibernate.annotations.CacheConcurrencyStrategy.NONSTRICT_READ_WRITE,
        include="non-lazy")
public class ChemicalStructure extends BaseEntity {

    @Id
    @GeneratedValue(strategy = GenerationType.AUTO,
            generator = "chem_structure_seq_gen")
    @SequenceGenerator(name = "chem_structure_seq_gen",
            sequenceName = "seq_chemical_structure", allocationSize = 1000)
    @Column(name = "structure_id")
    @Getter
    @Setter
    private Long id;

    @Column(name = "structure_key", unique = true, nullable=false, length = 1000)
    @Getter
    @Setter
    private String structureKey;

    @Column(name = "chemical_structure", nullable=false, columnDefinition = "TEXT")
    @Getter
    @Setter
    private String structureData;

    @Column(name = "molecular_weight", nullable=false)
    @Getter
    @Setter
    private double molecularWeight;

    @OneToMany(fetch = FetchType.LAZY, mappedBy = "pk.chemicalStructure")
    @OrderBy("pk.compound.id ASC")
    private List<ChemicalCompoundComposition> occurence = new ArrayList<>();

    @Transient
    @Getter
    @Setter
    private ChemistryFormat chemistryFormat;

    public ChemicalStructure() {
    }

    public ChemicalStructure(Long id, String structureKey,
            String structureData, double molecularWeight) {

        this.id = id;
        this.structureData = structureData;
        this.structureKey = structureKey;
        this.molecularWeight = molecularWeight;
    }

    public ChemicalStructure(String structureKey,
            String structureData, double molecularWeight) {
        this(null,structureKey, structureData, molecularWeight);
    }

    @JsonIgnore
    public String getMolfile() {
        if (chemistryFormat == ChemistryFormat.MOL){
            return structureData;
        }
        ChemicalFormatConverter converter = new ChemicalFormatConverter ();
        String molfile = converter.toMol(structureData);
        return molfile;
    }

    @JsonIgnore
    public String getMolfile(boolean aromatize) {
        if (chemistryFormat == ChemistryFormat.MOL){
            return structureData;
        }
        ChemicalFormatConverter converter = new ChemicalFormatConverter ();
        String molfile = converter.toMol(structureData, aromatize);
        return molfile;
    }

    @JsonIgnore
    public String getSmiles() {
        if (chemistryFormat == ChemistryFormat.SMI){
            return structureData;
        }
        ChemicalFormatConverter converter = new ChemicalFormatConverter ();
        String smiles = converter.toSmiles(structureData);
        return smiles;
    }

    /**
     * Sets all meta data fields including id to null.
     *
     */
    public void reset(){
        this.id = null;
        this.setCreated(null);
        this.setCreatedBy(null);
        this.setVersion(null);
    }
}
