/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.entity;

import com.fasterxml.jackson.annotation.JsonIdentityInfo;
import com.fasterxml.jackson.annotation.ObjectIdGenerators;
import java.util.HashSet;
import java.util.Set;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.ManyToMany;
import javax.persistence.SequenceGenerator;
import javax.persistence.Table;
import lombok.Data;
import lombok.EqualsAndHashCode;

/**
 * <p>
 * Defines a Role in your Application.
 * </p>
 * <p>
 * A role can be either a basic role (privilege) like "read_RegisttrationCompound"
 * that defines a specific privilege on a specific entity class. You can also create your own
 * roles but that requires a <code>RoleHierarchy</code> to be implemented. In a
 * Role hierarchy you define which Role is higher than anohter role and hence the
 * higher role automatically includes all lower roles. Eg. an actual role with
 * multiple privileges.
 * </p>
 * <p>
 * See <a href="https://bitbucket.org/kienerj/moleculedatabaseframework/wiki/Security">
 * Security Wiki for MoleculeDatabaseFramework</a> for further information.
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Entity
@Table(name="role_table")
@JsonIdentityInfo(generator=ObjectIdGenerators.IntSequenceGenerator.class, property="@id")
@Data
@EqualsAndHashCode(callSuper = false, of = {"roleName"})
public class Role extends BaseEntity {

    @Id
    @GeneratedValue(strategy = GenerationType.AUTO,
        generator = "role_id_seq_gen")
    @SequenceGenerator(name = "role_id_seq_gen", sequenceName = "seq_role")
    @Column(name = "role_id")
    private Long id;

    @Column(name="role_name", unique = true, nullable = false)
    private String roleName;

    @ManyToMany(fetch = FetchType.LAZY, mappedBy="roles",
        targetEntity=User.class, cascade = CascadeType.ALL)
    private Set<User> users = new HashSet<>();

    public Role(){}

    public Role(Long id, String roleName){
        this.id = id;
        this.roleName = roleName;
    }

    public Role(String roleName){
        this(null,roleName);
    }

    /**
     * Returns this <code>Roles roleName</code>.
     *
     * @return
     */
    @Override
    public String toString(){
        // if toString() is generated by lombok, leads to stackoverflowError
        return roleName;
    }
}
