/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.exceptionhandler;

import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.annotation.AfterThrowing;
import org.aspectj.lang.annotation.Aspect;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.BaseEntity;
import org.bitbucket.kienerj.moleculedatabaseframework.service.Service;
import org.bitbucket.kienerj.moleculedatabaseframework.service.UniquenesscheckResult;
import org.bitbucket.kienerj.moleculedatabaseframework.sql.UniqueConstraintViolationException;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.core.annotation.Order;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.stereotype.Component;

/**
 * Handles springs DataIntegrityViolationException
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Component
@Order(1500) // must be less than order of <tx:annotation-driven />
@Aspect
public class DIVExceptionHandler {

    private static final XLogger logger = XLoggerFactory.getXLogger("DIVExceptionHandler");

    @AfterThrowing(pointcut = "execution(* org.bitbucket.kienerj.moleculedatabaseframework.service.Service.save(*))", throwing = "ex")
    public void handleException(JoinPoint joinPoint, DataIntegrityViolationException ex) {

        logger.entry(joinPoint, ex);
        Service service = (Service)joinPoint.getTarget();
        // save() method always only has 1 argument
        BaseEntity offendingEntity = (BaseEntity)joinPoint.getArgs()[0];

        UniquenesscheckResult result = service.checkUniqueness(offendingEntity);
        if (!result.isUnique()) {
            String message = "Unique constraints for field(s): ";
            for (String fieldName : result.getViolatedContraints().keySet()) {
                // String concat in loop no issue as in probably 99.99% cases
                // there will only be 1 item in the collection.
                message += fieldName + ", ";
            }
            message = message.substring(0, message.length() - 2);
            message += " were violated.";
            UniqueConstraintViolationException uniqueEx =
                    new UniqueConstraintViolationException(message, result.getViolatedContraints(), ex);
            logger.throwing(uniqueEx);
            throw uniqueEx;
        } else {
            logger.throwing(ex);
            throw ex;
        }
    }
}
