/*
 * Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.io;

import java.text.ParseException;
import java.util.Map;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;

/**
 * Class for creating Containables from an SD-File.
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class ContainableFactory<T extends Containable> {

    private static final XLogger logger = XLoggerFactory.getXLogger("ContainableFactory");

    private final Class<T> containableClass;

    public ContainableFactory (Class<T> containableClass){
        this.containableClass = containableClass;
    }

    /**
     * <p>
     * Creates a new <code>Containable</code> Instance from the given property mappings
     * and property values.
     * </p>
     * <p>
     * This only sets "simple" properties like primitives, Strings or Dates.
     * Associations to ChemicalCompound and  ChemicalCompoundContainers are not set.
     * </p>
     * <p>
     * Date values must be in one of the following formats: yyyy-MM-dd HH:mm:ss<br>
     * yyyy-MM-dd<br> dd.mm.yyyy<br> mm/dd/yyy<br> where HH is the hour value
     * from 0 to 23 (see java.text.SimpleDateFormat)
     * </p>
     * @param mappedProperties
     * @return
     * @throws ParseException when the given data can not be converted to the target type
     * @throws IllegalAccessException
     * @throws NoSuchFieldException in case of an invalid mapping in <code>mappedProperties</code>
     * @throws InstantiationException if <code>compoundClass</code> does not have a zero argument constructor
     */
    public T createContainable(ChemicalCompound compound, Map<String, String> mappedProperties,
            Map<String, String> propertyValues) throws ParseException,
            IllegalAccessException, NoSuchFieldException, InstantiationException{

        logger.entry(mappedProperties, propertyValues);
        T containable = containableClass.newInstance();
        PropertySetter.setProperties(containable, containableClass, mappedProperties, propertyValues);
        containable.setChemicalCompound(compound);
        compound.addContainable(containable);
        logger.exit(containable);
        return containable;
    }
}
