/*
 * Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.io;

import java.lang.reflect.Field;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.util.ReflectionUtils;

/**
 * Util Class for setting "simple" properties in entity classes from imported
 * text files, namley SD-Files.
 *
 * "Simple" Properties -> any primitives, String and Date/Calendar fields.
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class PropertySetter {

    private static final XLogger logger = XLoggerFactory.getXLogger("PropertySetter");

    /**
     * <p>
     * Sets "simple" properties for a given <code>Object</code> from a mapping.
     * </p>
     * <p>
     * <code>mappedProperties</code> contains a mapping of source data property
     * name to a field existing in the given implementation of
     * <code>Object</code> or in a super class of it. Example:
     * </p>
     * <p>
     * <pre>
     * casNumber -> cas
     * name -> compoundName
     * </pre>
     * </p>
     * <p>
     * <code>properties</code> maps a source data property name to its value.
     * Example:
     * </p>
     * <pre>
     * casNumber -> 71-43-2
     * name -> benzene
     * </pre>
     *
     * <p>
     * Properties are mapped by name (case-sensitive) were the key
     * is the property in the source data and the value is the field name of the
     * entity.<br> <br> Type is resolved by the fields type.
     * Only primitives, Strings or dates can be imported. Date / Calendar values
     * must be in one of the following formats: yyyy-MM-dd HH:mm:ss<br>
     * yyyy-MM-dd<br> dd.mm.yyyy<br> mm/dd/yyy<br> where HH is the hour value
     * from 0 to 23 (see java.text.SimpleDateFormat)
     * </p>
     *
     * @param entity
     * @param mappedProperties
     * @param properties
     * @return
     * @throws ParseException
     * @throws IllegalAccessException
     * @throws NoSuchFieldException
     */
    public static void setProperties(Object entity, Class entityClass,
            Map<String, String> mappedProperties,
            Map<String, String> properties)
            throws ParseException, IllegalAccessException, NoSuchFieldException {

        logger.entry(entity, mappedProperties, properties);
        if (mappedProperties != null && properties != null) {
            for (Entry<String, String> entry : mappedProperties.entrySet()) {

                String fieldName = entry.getValue();

                //Get field recursively up the inheritance "tree"
                Field field = ReflectionUtils.findField(entityClass, fieldName);
                if (field == null) {
                    throw new NoSuchFieldException(String.format("No Field %s found in Class %s.", fieldName, entityClass.getSimpleName()));
                }
                setField(entity, field, properties.get(entry.getKey()));
            }
        }
        logger.exit();
    }

    private static void setField(Object entity,
            Field field, String value)
            throws ParseException, IllegalAccessException {

        Class fieldType = field.getType();
        logger.debug("Trying to set Field {} of type {} to " + value, field.getName(), fieldType.getName());
        field.setAccessible(true);
        if (fieldType.isPrimitive()) {

            if (fieldType.equals(Integer.TYPE)) {
                field.set(entity, Integer.parseInt(value));
            } else if (fieldType.equals(Long.TYPE)) {
                field.set(entity, Long.parseLong(value));
            } else if (fieldType.equals(Float.TYPE)) {
                field.set(entity, Float.parseFloat(value));
            } else if (fieldType.equals(Double.TYPE)) {
                field.set(entity, Double.parseDouble(value));
            } else if (fieldType.equals(Boolean.TYPE)) {
                field.set(entity, Boolean.parseBoolean(value));
            } else if (fieldType.equals(Short.TYPE)) {
                field.set(entity, Short.parseShort(value));
            } else if (fieldType.equals(Byte.TYPE)) {
                field.set(entity, Byte.parseByte(value));
            } else if (fieldType.equals(char.class)) {
                field.set(entity, value.charAt(0));
            }
        } else if (fieldType.equals(String.class)) {
            field.set(entity, value);

        } else if (fieldType.equals(Date.class)
                || fieldType.equals(Calendar.class)) {

            String dateFormatString = "";
            // yyyy-MM-dd
            final Pattern isoDateRegex = Pattern.compile(
                    "^(\\d{4})\\D?(0[1-9]|1[0-2])\\D?([12]\\d|0[1-9]|3[01])$");
            // yyyy-MM-dd HH:mm:ss
            final Pattern isoDateTimeRegex = Pattern.compile(
                    "^(\\d{4})\\D?(0[1-9]|1[0-2])\\D?([12]\\d|0[1-9]|3[01])(\\D?([01]\\d|2[0-3])\\D?([0-5]\\d)\\D?([0-5]\\d)?\\D?(\\d{3})?)?$");
            // dd.mm.yyyy
            final Pattern europeanDateRegex = Pattern.compile(
                    "^(0[1-9]|[12][0-9]|3[01])[.](0[1-9]|1[012])[.](19|20)\\d\\d$");
            // mm/dd/yyyy
            final Pattern usDateRegex = Pattern.compile(
                    "^(0[1-9]|1[012])[/](0[1-9]|[12][0-9]|3[01])[/](19|20)\\d\\d$");

            final Matcher isoDateMatcher = isoDateRegex.matcher(value);
            final Matcher isoDateTimeMatcher = isoDateTimeRegex.matcher(value);
            final Matcher europeanDateMatcher = europeanDateRegex.matcher(value);
            final Matcher usDateMatcher = usDateRegex.matcher(value);

            if (isoDateMatcher.matches()) {
                dateFormatString = "yyyy-MM-dd";
            } else if (isoDateTimeMatcher.matches()) {
                dateFormatString = "yyyy-MM-dd HH:mm:ss";
            } else if (europeanDateMatcher.matches()) {
                dateFormatString = "dd.MM.yyyy";
            } else if (usDateMatcher.matches()) {
                dateFormatString = "MM/dd/yyyy";
            }

            final SimpleDateFormat formatter =
                    new SimpleDateFormat(dateFormatString, Locale.getDefault());

            if (fieldType.equals(Date.class)) {
                final Date date = formatter.parse(value);
                field.set(entity, date);
            } else {
                final Calendar cal = Calendar.getInstance();
                cal.setTime(formatter.parse(value));
                field.set(entity, cal);
            }
        }
    }
}
