/*
 *  Copyright (C) 2011 Joos Kiener
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.io;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.ListIterator;
import java.util.NoSuchElementException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;

/**
 * <p>
 * Reads all entries (Molecules) and their properties and returns them as a
 * SdfRecord. The data is in raw text format as found in the file and might not
 * be a valid molecule.
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class RawSdfReader {

    private static final XLogger logger = XLoggerFactory.getXLogger("RawSdfReader");

    private BufferedReader bufferedReader;
    private SdfRecord nextRecord;
    private boolean isNextElementRead;
    private boolean hasNext;

    public RawSdfReader(Reader reader) {
        this.bufferedReader = new BufferedReader(reader);
        this.isNextElementRead = false;
        this.hasNext = true;
    }

    /**
     *
     *
     * @return whether a next record is available
     * @throws SdfImportException
     */
    public boolean hasNext() {
        logger.entry();
        try {
            if (!hasNext) {
                return false;
            }
            if (!isNextElementRead) {
                nextRecord = readNextRecord();
                isNextElementRead = true;
            }
            hasNext = nextRecord != null;
            logger.exit(hasNext);
            return hasNext;
        } catch (IOException ioEx) {
            throw new SdfImportException(ioEx);
        }
    }

    /**
     * Gets the next entry in the SD-File
     *
     * @return the next entry in the SD-File
     * @throws SdfImportException in case of an IO issue
     * @throws NoSuchElementException if EOF is reached
     */
    public SdfRecord next() {
        logger.entry();
        if (!hasNext) {
                throw new NoSuchElementException("Reached end of File.");
            }

        if (!isNextElementRead) {
            hasNext();
        }
        isNextElementRead = false;
        logger.exit(nextRecord);
        return nextRecord;
    }

    /**
     * Returns null if EOF is reached.
     *
     * @return the next element or null if EOF
     * @throws IOException
     */
    private SdfRecord readNextRecord() throws IOException {

        String lineToAdd;
        ArrayList<String> recordLines = new ArrayList<>();
        logger.debug("Start reading sdf record.");
        do {
            lineToAdd = bufferedReader.readLine();
            if (lineToAdd == null){
                return null;
            } else if (lineToAdd.equals("$$$$")){
                logger.debug("Found end of SDF Record.");
                break;
            }
            recordLines.add(lineToAdd);
        }
        while (true);

        boolean isFirstLine = true;
        StringBuilder molecule = new StringBuilder();
        SdfRecord sdfRecord = new SdfRecord();

        ListIterator<String> lineIterator = recordLines.listIterator();
        while (lineIterator.hasNext()) {
            String line = lineIterator.next();

            if (isFirstLine) {

                if (line.startsWith(">")) {
                    readSdfProperty(sdfRecord, lineIterator);
                } else {
                    sdfRecord.setMolfileName(line);
                    molecule.append(line);
                    molecule.append(System.getProperty("line.separator"));
                }
                isFirstLine = false;
            } else if (line.startsWith(">")) {
                readSdfProperty(sdfRecord, lineIterator);
            } else {
                molecule.append(line);
                molecule.append(System.getProperty("line.separator"));
            }
        }
        sdfRecord.setMolfile(molecule.toString());
        return sdfRecord;
    }

    private void readSdfProperty(SdfRecord sdfRecord,
            ListIterator<String> lineIterator) {

        logger.debug("Parsing property...");
        String line = lineIterator.previous();
        lineIterator.next();
        // used to read property names in SD-File which are between <>
        Pattern regex = Pattern.compile("(?<=\\<)(.*?)(?=\\>)");

        Matcher matcher = regex.matcher(line);
        matcher.find();
        String propertyName = matcher.group();
        // value of property is on next line
        String value = "";
        while (lineIterator.hasNext()) {
            line = lineIterator.next();
            if (line.equals("")) {
                break;
            } else if (line.startsWith(">")) {
                lineIterator.previous();
                break;
            } else {
                // String concat in loop is no issue here. In most cases
                // there will only be 1 line.
                value = value + System.getProperty("line.separator") + line;
            }
        }
        // remove possible leading line.separator
        value = value.trim();
        sdfRecord.addProperty(propertyName, value);
        logger.debug("Adding property {} with value {}.", propertyName, value);
    }
}
