/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.io;

import com.ggasoftware.indigo.Indigo;
import com.ggasoftware.indigo.IndigoObject;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.Map;
import java.util.Set;
import lombok.AllArgsConstructor;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.bitbucket.kienerj.moleculedatabaseframework.service.ChemicalCompoundService;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.transaction.annotation.Transactional;

/**
 * <p> Exports a List of compounds into an SD-File. </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@AllArgsConstructor
public class SdfCompoundExporter<T extends ChemicalCompound<? extends Containable>> {

    private static final XLogger logger = XLoggerFactory.getXLogger("SdfCompoundExporter");
    private static final String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";
    private static final String FILE_PREFIX = "MDF-Export";
    private static final String FILE_SUFFIX = ".sdf";
    // if user requests to export more than below amount of compounds
    // the sd-file is generated on the server else it is generated in-memory.
    private static final int EXPORT_SIZE_THRESHHOLD = 1000;
    private ChemicalCompoundService<T> compoundService;

    /**
     * <p> Exports the given List of compounds using the given writer into an
     * SD-File. </p> <p> If the compound contains of multiple structures, the
     * percentages are exported as a property of name structureKey and the value
     * is the percentage for that structure. </p>
     *
     * <p> The
     * <code>Collection<String> propertiesToExport</code> must contain the names
     * of all the fields you want to export. If it is empty, only the structure
     * will be exported, if it is null, all fields will be exported. </p> <p>
     * For date fields default format of "yyyy-MM-dd HH:mm:ss" is used. </p>
     *
     * @param compoundIds
     * @param writer
     */
    @Transactional
    public void exportCompoundsToSdf(Collection<Long> compoundIds, Writer writer,
            Set<String> propertiesToExport) {

        exportCompoundsToSdf(compoundIds, writer, propertiesToExport, DATE_FORMAT);
    }

    /**
     * <p> Exports the given List of compounds using the given writer into an
     * SD-File. </p> <p> If the compound contains of multiple structures, the
     * percentages are exported as a property of name structureKey and the value
     * is the percentage for that structure. </p>
     *
     * <p> The
     * <code>Collection<String> propertiesToExport</code> must contain the names
     * of all the fields you want to export. If it is empty, only the structure
     * will be exported, if it is null, all fields will be exported. </p> <p>
     * For date fields the date format defined by
     * <code>dateFormat</code> is used. See <a
     * href="http://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html">SimpleDateFormat</a>
     * for more information. </p>
     *
     * @param compoundIds
     * @param writer
     */
    @Transactional
    public void exportCompoundsToSdf(Collection<Long> compoundIds, Writer writer,
            Set<String> propertiesToExport, String dateFormat) {

        logger.entry(compoundIds, writer, propertiesToExport, dateFormat);
        /*
         * If true writes sd-file on server and then streams it to client else
         * sd-file is generated in memory and streamed to client.
         */
        boolean useTempfile = false;
        File tempFile = null;
        if (compoundIds.size() > EXPORT_SIZE_THRESHHOLD) {
            useTempfile = true;
            try {
                tempFile = File.createTempFile(FILE_PREFIX, FILE_SUFFIX);
            } catch (IOException ex) {
                useTempfile = false;
            }
        }
        Indigo indigo = new Indigo();
        IndigoObject buf;
        if (useTempfile) {
            buf = indigo.writeFile(tempFile.getAbsolutePath());
        } else {
            buf = indigo.writeBuffer();
        }

        try {

            for (Long id : compoundIds) {

                T compound = compoundService.getById(id);

                IndigoObject mol = compound.getIndigoMolecule(indigo);

                logger.debug("Setting properties for SDF-Export...");
                Map<String, String> compoundProperties = PropertyGetter
                        .getProperties(compound, compound.getClass(),
                        propertiesToExport, dateFormat);

                for (Map.Entry<String, String> entry : compoundProperties.entrySet()) {
                    mol.setProperty(entry.getKey(), entry.getValue());
                }

                mol.setName(compound.getCompoundName());

                logger.debug("Writing Compound to SD-File...");
                buf.sdfAppend(mol);
            }
            if (useTempfile) {
                buf.close();
                try (BufferedReader reader = new BufferedReader(new FileReader(tempFile))) {
                    String line;
                    while ((line = reader.readLine()) != null) {
                        writer.write(line);
                        writer.write(System.getProperty("line.separator"));
                    }
                }
                logger.exit();
            } else {
                writer.write(buf.toString());
                logger.exit();
            }
        } catch (IOException | NoSuchFieldException ex) {
            throw new SdfExportException(ex);
        } finally {
            if (useTempfile && buf != null) {
                buf.close();
            }
            if (tempFile != null) {
                tempFile.delete();
            }
        }
    }
}
