/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.io;

import com.ggasoftware.indigo.Indigo;
import com.ggasoftware.indigo.IndigoObject;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import lombok.AllArgsConstructor;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundContainer;
import org.bitbucket.kienerj.moleculedatabaseframework.service.CompoundContainerService;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.transaction.annotation.Transactional;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@AllArgsConstructor
public class SdfContainerExporter<T extends ChemicalCompoundContainer> {

    private static final XLogger logger = XLoggerFactory.getXLogger("SdfContainerExporter");
    private static final String FILE_PREFIX = "MDF-Export";
    private static final String FILE_SUFFIX = ".sdf";
    // if user requests to export more than below amount of compounds
    // the sd-file is generated on the server else it is generated in-memory.
    private static final int EXPORT_SIZE_THRESHHOLD = 1000;
    private static final String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";
    private final CompoundContainerService<T> containerService;

    /**
     * <p> Exports the given List of container using the given writer into an
     * SD-File. </p> <p> If the compound in the container consists of multiple
     * structures, the percentages are exported as a property of name inChikey
     * and the value is the percentage for that structure. </p> * <p> The
     * <code>Sets<String></code> must contain the names of all the fields you
     * want to export for the according entity. If it is empty, only the
     * structure will be exported, if it is null, all fields will be exported.
     * </p> <p> For date fields default format of "yyyy-MM-dd HH:mm:ss" is used.
     * </p>
     *
     * @param compoundIds
     * @param writer
     * @throws SdfExportException
     */
    @Transactional
    public void exportContainersToSdf(Collection<Long> containerIds, Writer writer,
            Set<String> containerPropertyNames,
            Set<String> containablePropertyNames,
            Set<String> compoundPropertyNames) {

        exportContainersToSdf(containerIds, writer, containerPropertyNames,
                containablePropertyNames, compoundPropertyNames, DATE_FORMAT);
    }

    /**
     * <p> Exports the given List of container using the given writer into an
     * SD-File. </p> <p> If the compound in the container consists of multiple
     * structures, the percentages are exported as a property of name inChikey
     * and the value is the percentage for that structure. </p> * <p> The
     * <code>Sets<String></code> must contain the names of all the fields you
     * want to export for the according entity. If it is empty, only the
     * structure will be exported, if it is null, all fields will be exported.
     * </p> <p> For date fields the date format defined by
     * <code>dateFormat</code> is used. See <a
     * href="http://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html">SimpleDateFormat</a>
     * for more information. </p>
     *
     * @param compoundIds
     * @param writer
     * @param dateForm the date format to use for date fields.
     * @throws SdfExportException
     */
    @Transactional
    public void exportContainersToSdf(Collection<Long> containerIds, Writer writer,
            Set<String> containerPropertyNames,
            Set<String> containablePropertyNames,
            Set<String> compoundPropertyNames,
            String dateFormat) {

        logger.entry(containerIds, writer, containerPropertyNames,
                containablePropertyNames, compoundPropertyNames, dateFormat);
        /*
         * If true writes sd-file on server and then streams it to client else
         * sd-file is generated in memory and streamed to client.
         */
        boolean useTempfile = false;
        File tempFile = null;
        if (containerIds.size() > EXPORT_SIZE_THRESHHOLD) {
            useTempfile = true;
            try {
                tempFile = File.createTempFile(FILE_PREFIX, FILE_SUFFIX);
            } catch (IOException ex) {
                useTempfile = false;
            }
        }
        Indigo indigo = new Indigo();
        IndigoObject buf;
        if (useTempfile) {
            logger.debug("Using temp file as buffer...");
            buf = indigo.writeFile(tempFile.getAbsolutePath());
        } else {
            logger.debug("Using memory as buffer...");
            buf = indigo.writeBuffer();
        }

        try {

            for (Long id : containerIds) {

                T container = containerService.getById(id);

                IndigoObject mol = container.getChemicalCompound().getIndigoMolecule(indigo);

                logger.debug("Setting properties for SDF-Export...");
                Map<String, String> allProperties = new HashMap<>();

                Map<String, String> containerProperties = PropertyGetter
                        .getProperties(container, container.getClass(),
                        containerPropertyNames, dateFormat);

                Map<String, String> containableProperties = PropertyGetter
                        .getProperties(container.getContainable(),
                        container.getContainable().getClass(),
                        containablePropertyNames, dateFormat);

                Map<String, String> compoundProperties = PropertyGetter
                        .getProperties(container.getChemicalCompound(),
                        container.getChemicalCompound().getClass(),
                        compoundPropertyNames, dateFormat);

                allProperties.putAll(containerProperties);
                allProperties.putAll(containableProperties);
                allProperties.putAll(compoundProperties);

                for (Map.Entry<String, String> entry : allProperties.entrySet()) {
                    mol.setProperty(entry.getKey(), entry.getValue());
                }
                mol.setName(container.getChemicalCompound().getCompoundName());
                logger.debug("Writing Container to SD-File...");
                buf.sdfAppend(mol);
            }
            if (useTempfile) {
                buf.close();
                logger.debug("Reading temp file into Writer...");
                try (BufferedReader reader = new BufferedReader(new FileReader(tempFile))) {
                    String line;
                    while ((line = reader.readLine()) != null) {
                        writer.write(line);
                        writer.write(System.getProperty("line.separator"));
                    }
                }
                logger.exit();
            } else {
                logger.debug("Sending buffer to Writer...");
                writer.write(buf.toString());
                logger.exit();
            }
        } catch (IOException | NoSuchFieldException ex) {
            throw new SdfExportException(ex);
        } finally {
            if (useTempfile && buf != null) {
                buf.close();
            }
            if (tempFile != null) {
                tempFile.delete();
            }
        }
    }
}
