/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.repository;

import com.mysema.query.jpa.impl.JPAQuery;
import com.mysema.query.types.ConstantImpl;
import com.mysema.query.types.Expression;
import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.expr.BooleanExpression;
import com.mysema.query.types.path.NumberPath;
import com.mysema.query.types.path.PathBuilder;
import com.mysema.query.types.template.BooleanTemplate;
import java.util.List;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.BaseEntity;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.QChemicalStructure;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.StructureSearchType;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cache.Cache;
import org.springframework.cache.CacheManager;
import org.springframework.data.domain.Sort;
import org.springframework.data.repository.NoRepositoryBean;

/**
 * <p>Abstract implementation of <code>ChemicalStructureSearchRepository</code> that
 * contains shared fields and methods.</p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@NoRepositoryBean
public abstract class AbstractStructureSearchRepositoryImpl<T extends BaseEntity>
        implements ChemicalStructureSearchRepository<T>, IdRepository<T> {

    /*
     * Note: This class does not implement findByChemicalStructure and
     * findBySimilarStructure methods because @Cachebale annotation requires
     * a constant string Expression but each implementation should have a separate
     * string, eg. use a separate cache. Why? Because if we add a new Container
     * we must clear that search cache. However the search cache for compounds
     * is still valid and must not be cleared.
     */

    private static final XLogger logger = XLoggerFactory.getXLogger("AbstractStructureSearchRepository");
    @PersistenceContext
    protected EntityManager entityManager;
    @Autowired
    protected CacheManager cacheManager;
    // default if property not set: true
    @Value("${structureSearch.selectDistinct:true}")
    protected boolean selectDistinct;
    // default if property not set: 12
    @Value("${structureSearch.useIndexThreshold:12}")
    protected long countDifferenceTH;
    protected final String cacheName;

    public AbstractStructureSearchRepositoryImpl(String cacheName) {
        this.cacheName = cacheName;
    }

    public AbstractStructureSearchRepositoryImpl(EntityManager entityManager,
            CacheManager cacheManager, String cacheName) {
        this.entityManager = entityManager;
        this.cacheManager = cacheManager;
        this.cacheName = cacheName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findAll(Predicate predicate,
            PathBuilder<T> pathBuilder) {
        logger.entry(predicate);

        JPAQuery query = new JPAQuery(entityManager);
        query = query.from(pathBuilder)
                .where(predicate);
        NumberPath<Long> idPath = new NumberPath<>(Long.class, pathBuilder, "id");
        List<Long> ids = query.distinct().list(pathBuilder.get(idPath));
        logger.exit(ids);
        return ids;
    }

    protected BooleanExpression buildStructureSearchExpression(
            String structureData, StructureSearchType searchType,
            Predicate predicate, String searchOptions,
            long countDifferenceFactor) {

        BooleanExpression matchesStructureQuery;
        QChemicalStructure structure = QChemicalStructure.chemicalStructure;

        switch (searchType) {
            case EXACT:
                //always use index
                matchesStructureQuery = BooleanTemplate.create(
                        "isExactStructure({0},{1},{2}) = true",
                        structure.structureData,
                        ConstantImpl.create(structureData),
                        ConstantImpl.create(searchOptions));
                break;
            case SUBSTRUCTURE:
                if (countDifferenceFactor < countDifferenceTH) {
                    matchesStructureQuery = BooleanTemplate.create(
                            "isSubstructure({0},{1},{2}) = true",
                            structure.structureData,
                            ConstantImpl.create(structureData),
                            ConstantImpl.create(searchOptions));
                } else {
                    matchesStructureQuery = BooleanTemplate.create(
                            "matchSub({0},{1},{2}) = true",
                            structure.structureData,
                            ConstantImpl.create(structureData),
                            ConstantImpl.create(searchOptions));
                }
                break;
            case SMARTS:
                if (countDifferenceFactor < countDifferenceTH) {
                    matchesStructureQuery = BooleanTemplate.create(
                            "matchesSMARTS({0},{1},{2}) = true",
                            structure.structureData,
                            ConstantImpl.create(structureData),
                            ConstantImpl.create(searchOptions));
                } else {
                    matchesStructureQuery = BooleanTemplate.create(
                            "matchSmarts({0},{1},{2}) = true",
                            structure.structureData,
                            ConstantImpl.create(structureData),
                            ConstantImpl.create(searchOptions));
                }
                break;
            case FORMULA:
                if (structureData.startsWith("=")
                        || structureData.startsWith("<=")) {
                    matchesStructureQuery = BooleanTemplate.create(
                            "matchesFormula({0},{1},{2}) = true",
                            structure.structureData,
                            ConstantImpl.create(structureData),
                            ConstantImpl.create(searchOptions));
                } else if (structureData.startsWith(">=")) {
                    // don't use index due to bug in bingo
                    matchesStructureQuery = BooleanTemplate.create(
                            "matchgross({0},{1},{2}) = true",
                            structure.structureData,
                            ConstantImpl.create(structureData),
                            ConstantImpl.create(searchOptions));
                } else {
                    throw new IllegalArgumentException("Formula Query must start with '=', '<=' or '>='.");
                }
                break;
            default:
                throw new IllegalArgumentException("Invalid search type.");
        }
        return matchesStructureQuery;
    }

    protected long getCount(JPAQuery query,
            String pathName) {
        logger.debug("Getting total result count for query {}", query.toString());
        Predicate whereClause = query.getMetadata().getWhere();
        String key = whereClause.toString() + pathName;
        Cache cache = this.cacheManager.getCache(cacheName);
        Cache.ValueWrapper value = cache.get(key);
        if (value == null) {
            Long result = query.distinct().count();
            cache.put(key, result);
            return result;
        }
        return (Long) value.get();
    }

    protected long getTotalStructureCount() {
        logger.debug("Getting total structure count...");
        Cache cache = this.cacheManager.getCache(cacheName);
        String key = "totalStructureCount";
        Cache.ValueWrapper value = cache.get(key);
        if (value == null) {
            QChemicalStructure structure = QChemicalStructure.chemicalStructure;
            JPAQuery countStructuresQuery = new JPAQuery(entityManager);
            long totalStructureCount = countStructuresQuery.from(structure).count();
            cache.put(key, totalStructureCount);
            return totalStructureCount;
        } else {
            return (Long) value.get();
        }
    }

    /**
     * Modified method from
     * org.springframework.data.jpa.repository.support.Querydsl
     *
     */
    protected JPAQuery applySorting(Sort sort, JPAQuery query,
            PathBuilder<T> pathBuilder) {

        if (sort == null) {
            return query;
        }

        for (Sort.Order order : sort) {
            query.orderBy(toOrder(order, pathBuilder));
        }

        return query;
    }

    /**
     * Modified method from
     * org.springframework.data.jpa.repository.support.Querydsl
     *
     */
    @SuppressWarnings({"rawtypes", "unchecked"})
    private OrderSpecifier<?> toOrder(Sort.Order order,
            PathBuilder<T> pathBuilder) {

        Expression<Object> property = pathBuilder.get(order.getProperty());

        return new OrderSpecifier(order.isAscending() ? com.mysema.query.types.Order.ASC
                : com.mysema.query.types.Order.DESC, property);
    }
}
