/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.repository;

import com.google.common.base.Preconditions;
import com.mysema.query.jpa.impl.JPAQuery;
import com.mysema.query.types.ConstantImpl;
import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.expr.BooleanExpression;
import com.mysema.query.types.path.NumberPath;
import com.mysema.query.types.path.PathBuilder;
import com.mysema.query.types.template.BooleanTemplate;
import com.mysema.query.types.template.NumberTemplate;
import java.util.List;
import javax.persistence.EntityManager;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundContainer;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.QChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.QChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.QChemicalCompoundContainer;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.QChemicalStructure;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.QContainable;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.SimilarityType;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.StructureSearchType;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.cache.CacheManager;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;

/**
 * <p> This class supplies chemical structure search for repositories for
 * classes extending
 * <code>ChemicalCompoundContainer</code>. </p> <p> Usage:<br> <br> Create an
 * interface that extends
 * <code>ChemicalStructureSearchRepository<T></code>,
 * <code>JpaRepository<T, Long></code> and
 * <code>QueryDslPredicateExecutor<T></code> <br> T = Type of class extending
 * AbstractChemcialContainer <br> Then extend this class an name it
 * InterfaceNameImpl. </p>
 *
 * <p> Spring Data automatically adds the implemented custom methods to
 * <code>ChemicalCompoundContainerRepository</code>. See <a
 * href="http://static.springsource.org/spring-data/data-jpa/docs/current/reference/html/repositories.html#repositories.custom-implementations">Spring
 * Data JPA documentation</a> for further information. </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class CompoundContainerRepositoryImpl<T extends ChemicalCompoundContainer>
        extends AbstractStructureSearchRepositoryImpl<T> {

    private static final XLogger logger = XLoggerFactory.getXLogger("CompoundContainerRepository");
    private static final String STRUCTURE_QUERY_CACHE = "containerStructureQuery";

    /**
     * For JPA Provider
     */
    public CompoundContainerRepositoryImpl() {
        super(STRUCTURE_QUERY_CACHE);
    }

    public CompoundContainerRepositoryImpl(EntityManager entityManager,
            CacheManager cacheManager) {
        super(entityManager, cacheManager, STRUCTURE_QUERY_CACHE);
    }

    /**
     * {@inheritDoc}
     */
    @Cacheable(STRUCTURE_QUERY_CACHE)
    @Override
    public Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Pageable pageable,
            Predicate predicate, String searchOptions,
            PathBuilder<T> pathBuilder) {

        logger.entry(structureData, searchType, pageable, predicate, searchOptions,
                pathBuilder);

        Preconditions.checkArgument(pageable != null);

        JPAQuery baseQuery = buildStructureQuery(structureData, searchType,
                predicate, searchOptions, pathBuilder);

        long count = getCount(baseQuery, pathBuilder.toString());

        JPAQuery finalQuery = baseQuery.limit(pageable.getPageSize())
                .offset(pageable.getOffset());

        finalQuery = applySorting(pageable.getSort(), finalQuery,
                pathBuilder);

        logger.debug("Executing chemical structure Query...");
        List<T> result;
        if (selectDistinct) {
            result = finalQuery.distinct().list(pathBuilder);
        } else {
            result = finalQuery.list(pathBuilder);
        }

        PageImpl<T> page = new PageImpl(result, pageable, count);
        logger.exit(page);
        return page;
    }

    @Cacheable(STRUCTURE_QUERY_CACHE)
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate, String searchOptions,
            PathBuilder<T> containerPathBuilder, OrderSpecifier<?>... orders) {

        logger.entry(structureData, searchType, predicate, searchOptions,
                containerPathBuilder, orders);

        JPAQuery jpaQuery = buildStructureQuery(structureData, searchType, predicate,
                searchOptions, containerPathBuilder);

        jpaQuery = jpaQuery.orderBy(orders);
        NumberPath<Long> idPath = new NumberPath<>(Long.class, containerPathBuilder, "id");
        List<Long> ids = jpaQuery.distinct().list(containerPathBuilder.get(idPath));
        logger.exit(ids);
        return ids;
    }

    /**
     * {@inheritDoc}
     */
    @Cacheable(STRUCTURE_QUERY_CACHE)
    @Override
    public Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType, Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate,
            PathBuilder<T> pathBuilder) {

        logger.entry(structureData, similarityType, lowerBound, upperBound,
                pageable, predicate, pathBuilder);

        Preconditions.checkArgument(pageable != null);

        JPAQuery baseQuery = this.buildSimilarityQuery(structureData,
                similarityType, lowerBound, upperBound, predicate, pathBuilder);

        long count = getCount(baseQuery, pathBuilder.toString());

        JPAQuery finalQuery = baseQuery.limit(pageable.getPageSize())
                .offset(pageable.getOffset());

        finalQuery = applySorting(pageable.getSort(), finalQuery,
                pathBuilder);

        logger.debug("Executing Similarity Query...");
        List<T> result = finalQuery.distinct().list(pathBuilder);

        PageImpl<T> page = new PageImpl(result, pageable, count);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Cacheable(STRUCTURE_QUERY_CACHE)
    @Override
    public List<Long> findBySimilarStructure(String structureData,
            SimilarityType similarityType, Double lowerBound, Double upperBound,
            Predicate predicate, PathBuilder<T> pathBuilder,
            OrderSpecifier<?>... orders) {

        logger.entry(structureData, similarityType, lowerBound, upperBound,
                predicate, pathBuilder, orders);

        JPAQuery jpaQuery = buildSimilarityQuery(structureData, similarityType,
                lowerBound, upperBound, predicate, pathBuilder);

        jpaQuery = jpaQuery.orderBy(orders);
        NumberPath<Long> idPath = new NumberPath<>(Long.class, pathBuilder, "id");
        List<Long> ids = jpaQuery.distinct().list(pathBuilder.get(idPath));
        logger.exit(ids);
        return ids;
    }

    private JPAQuery buildStructureQuery(String structureData,
            StructureSearchType searchType, Predicate predicate,
            String searchOptions, PathBuilder<T> pathBuilder) {

        Preconditions.checkArgument(structureData != null);
        Preconditions.checkArgument(searchType != null);
        Preconditions.checkArgument(pathBuilder != null);

        JPAQuery baseQuery = new JPAQuery(entityManager);

        QChemicalCompoundContainer basicContainer =
                new QChemicalCompoundContainer(pathBuilder);

        QContainable containable = QContainable.containable;
        QChemicalCompound compound = QChemicalCompound.chemicalCompound;
        QChemicalCompoundComposition composition =
                QChemicalCompoundComposition.chemicalCompoundComposition;
        QChemicalStructure structure = QChemicalStructure.chemicalStructure;

        /*
         * Bingo for PostgreSQL currently has no cost estimation. It is hard-coded
         * and structure index scanning is greatly underestimated. If a query has
         * additional filters like molecular weight or compound name, it can
         * be faster to use the bingo.matchsub function for filtering instead
         * of using the structure index and filter on compound name. The bigger the
         * database the bigger the effect of this issue.
         *
         * Hence the application code decides if index or function is used. If
         * predicate is null, the index will always be used. Else the ratio of
         * structures over hits found by filtering with the predictae is determined.
         * If all compounds consist of a single chemical structure the index scan
         * is about 12x times faster. So if the total amount of structures is over 12
         * times bigger than the result size of the predicate-query, the matchsub
         * function is used instead of the index. This ratio however greatly
         * depends on the data in the database!
         *
         */

        long countDifferenceFactor = 0;

        if (predicate != null) {
            JPAQuery countQuery = new JPAQuery(entityManager);

            countQuery = countQuery.from(pathBuilder)
                    .innerJoin(basicContainer.containable, containable)
                    .innerJoin(containable.chemicalCompound, compound)
                    .innerJoin(compound.compositions, composition)
                    .innerJoin(composition.pk.chemicalStructure, structure)
                    .where(predicate);

            long predicateCount = getCount(countQuery, pathBuilder.toString());
            long totalStructures = getTotalStructureCount();
            if (predicateCount != 0) {
                countDifferenceFactor = totalStructures / predicateCount;
            } else {
                countDifferenceFactor = countDifferenceTH + 1L; // don't use structure index
            }
        }

        BooleanExpression matchesStructureQuery = buildStructureSearchExpression(
                structureData, searchType, predicate, searchOptions, countDifferenceFactor);

        baseQuery = baseQuery.from(pathBuilder)
                .innerJoin(basicContainer.containable, containable)
                .innerJoin(containable.chemicalCompound, compound)
                .innerJoin(compound.compositions, composition)
                .innerJoin(composition.pk.chemicalStructure, structure)
                .where(matchesStructureQuery.and(predicate));

        return baseQuery;
    }

    private JPAQuery buildSimilarityQuery(String structureData,
            SimilarityType similarityType, Double lowerBound, Double upperBound,
            Predicate predicate, PathBuilder<T> pathBuilder) {

        Preconditions.checkArgument(structureData != null);
        Preconditions.checkArgument(pathBuilder != null);
        Preconditions.checkArgument(similarityType != null);
        Preconditions.checkArgument(lowerBound != null);

        // default upper bound
        if (upperBound == null) {
            upperBound = 1.0;
        }

        JPAQuery baseQuery = new JPAQuery(entityManager);

        QChemicalCompoundContainer basicContainer =
                new QChemicalCompoundContainer(pathBuilder);

        QContainable containable = QContainable.containable;
        QChemicalCompound compound = QChemicalCompound.chemicalCompound;
        QChemicalCompoundComposition composition =
                QChemicalCompoundComposition.chemicalCompoundComposition;
        QChemicalStructure structure = QChemicalStructure.chemicalStructure;

        BooleanExpression isSimilarStructure = BooleanTemplate.create(
                "isSimilarStructure({0},{1},{2},{3},{4}) = true",
                structure.structureData,
                NumberTemplate.create(Double.class, lowerBound.toString()),
                NumberTemplate.create(Double.class, upperBound.toString()),
                ConstantImpl.create(structureData),
                ConstantImpl.create(similarityType.getMetric()));

        baseQuery = baseQuery.from(pathBuilder)
                .innerJoin(basicContainer.containable, containable)
                .innerJoin(containable.chemicalCompound, compound)
                .innerJoin(compound.compositions, composition)
                .innerJoin(composition.pk.chemicalStructure, structure)
                .where(isSimilarStructure.and(predicate));
        return baseQuery;
    }
}
