/*
 * Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.security;

import java.util.Collection;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundContainer;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.hierarchicalroles.RoleHierarchy;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.userdetails.UserDetails;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class ChemicalCompoundContainerPermission extends AbstractPermission {

    @Autowired
    private RoleHierarchy roleHierarchy;
    private static final String INVALID_PERMISSION_MESSAGE = "%s is not a valid permission type. "
            + "Permission Type must be 'save' or 'delete'.";

    public ChemicalCompoundContainerPermission() {
        super("ChemicalCompoundContainer");
    }

    @Override
    public boolean isAllowed(Authentication authentication, Object targetDomainObject,
            String permissionType, String permissionTarget) {

        if (!isAuthenticated(authentication)) {
            return false;
        }

        boolean hasPermission = false;
        switch (permissionType) {
            case "save":
                hasPermission = canSave(authentication,
                        (ChemicalCompoundContainer) targetDomainObject, permissionTarget);
                break;
            case "delete":
                hasPermission = canDelete(authentication,
                        (ChemicalCompoundContainer) targetDomainObject, permissionTarget);
                break;
            default:
                throw new InvalidPermissionException(String.format(INVALID_PERMISSION_MESSAGE, permissionType));
        }
        return hasPermission;
    }

    private boolean canSave(final Authentication authentication,
            final ChemicalCompoundContainer targetDomainObject,
            final String permissionTarget) {

        final boolean isNewContainer = targetDomainObject.getId() == null;

        final boolean canSave = (hasRole(authentication, "create_" + permissionTarget) && isNewContainer)
                || (hasRole(authentication, "update_created_" + permissionTarget)
                && isCreator(authentication, targetDomainObject))
                || hasRole(authentication, "update_" + permissionTarget);

        if (!canSave) {
            return false;
        }

        /*
         * Relationship of entities have CascadeType.PERSIST set but not .MERGE hence
         * if targetDomainObject is a new entity we need to check if current
         * user has privilege to create any of the associated entites in targetDomainObject.
         * associated entites -> Containable and Compound in Containable
         *
         * Since CascadeType.MERGE is not set, changes to associated entites
         * are ignored when updating and hence no need to check if user can update them
         * (Checking if entity has changed is IMHO notn-trivial here in this context
         * hence this design)
         */

        final boolean isNewContainable = targetDomainObject.getContainable().getId() == null;
        final boolean isNewCompound = targetDomainObject.getChemicalCompound().getId() == null;

        if (isNewContainable
                && !canCreateContainable(authentication, targetDomainObject.getContainable())) {
            return false;
        }

        if (isNewCompound
                && !canCreateCompound(authentication, targetDomainObject.getChemicalCompound())) {
            return false;
        }

        return canSave;
    }

    private boolean canDelete(final Authentication authentication,
            final ChemicalCompoundContainer targetDomainObject,
            final String permissionTarget) {
        return (hasRole(authentication, "delete_created_" + permissionTarget)
                && isCreator(authentication, targetDomainObject))
                || hasRole(authentication, "delete_" + permissionTarget);
    }

    private boolean canCreateContainable(final Authentication authentication,
            final Containable containable) {

        final boolean isNewContainable = containable.getId() == null;
        final String containableClassName = containable.getClass().getSimpleName();
        return (hasRole(authentication, "create_" + containableClassName) && isNewContainable);
    }

    private boolean canCreateCompound(final Authentication authentication,
            final ChemicalCompound compound) {

        final boolean isNewContainable = compound.getId() == null;
        final String compoundClassName = compound.getClass().getSimpleName();
        return (hasRole(authentication, "create_" + compoundClassName) && isNewContainable);
    }

    private boolean hasRole(final Authentication authentication,
            final String role) {
        final Collection<? extends GrantedAuthority> authorities =
                roleHierarchy.getReachableGrantedAuthorities(authentication.getAuthorities());
        for (GrantedAuthority authoritiy : authorities) {
            if (authoritiy.getAuthority().equals(role)) {
                return true;
            }
        }
        return false;
    }

    private boolean isCreator(final Authentication authentication,
            final Object targetDomainObject) {

        final String userName = getLogin(authentication);
        return ((ChemicalCompoundContainer) targetDomainObject).getCreatedBy().equals(userName);
    }

    private String getLogin(final Authentication authentication) {
        final Object authenticationUserDetails = authentication.getPrincipal();
        return ((UserDetails) authenticationUserDetails).getUsername();
    }

    private boolean isAuthenticated(final Authentication authentication) {
        return authentication != null && authentication.getPrincipal() instanceof UserDetails;
    }
}
