/*
 * Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.security;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundContainer;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.hierarchicalroles.RoleHierarchy;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.userdetails.UserDetails;

/**
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class ContainablePermission extends AbstractPermission {

    @Autowired
    private RoleHierarchy roleHierarchy;
    private static final String INVALID_PERMISSION_MESSAGE = "%s is not a valid permission type. "
            + "Permission Type must be 'save' or 'delete'.";

    public ContainablePermission() {
        super("Containable");
    }

    @Override
    public boolean isAllowed(Authentication authentication, Object targetDomainObject,
            String permissionType, String permissionTarget) {

        if (!isAuthenticated(authentication)) {
            return false;
        }

        boolean hasPermission = false;
        switch (permissionType) {
            case "save":
                hasPermission = canSave(authentication, (Containable) targetDomainObject, permissionTarget);
                break;
            case "delete":
                hasPermission = canDelete(authentication, (Containable) targetDomainObject, permissionTarget);
                break;
            default:
                throw new InvalidPermissionException(String.format(INVALID_PERMISSION_MESSAGE, permissionType));
        }
        return hasPermission;
    }

    private boolean canSave(Authentication authentication, Containable targetDomainObject,
            String permissionTarget) {

        boolean isNew = targetDomainObject.getId() == null;

        boolean canSave = (hasRole(authentication, "create_" + permissionTarget) && isNew)
                || (hasRole(authentication, "update_created_" + permissionTarget)
                && isCreator(authentication, targetDomainObject))
                || hasRole(authentication, "update_" + permissionTarget);

        if (!canSave){
            return false;
        }

        /*
         * Relationship of entities have CascadeType.PERSIST set but not .MERGE hence
         * if targetDomainObject is a new entity we need to check if current
         * user has privilege to create any of the associated entites in targetDomainObject.
         * associated entites -> Containers and Compound
         *
         * Since CascadeType.MERGE is not set, changes to associated entites
         * are ignored when updating and hence no need to check if user can update them
         * (Checking if entity has changed is IMHO notn-trivial here in this context
         * hence this design)
         */

        boolean isNewCompound = targetDomainObject.getChemicalCompound().getId() == null;

        // compound can already exist. If compound is new, check if user can
        // create a new compound of this type
        if(isNew && isNewCompound){
            if(!canCreateCompound(authentication, targetDomainObject.getChemicalCompound())){
                return false;
            }
        }
        if(isNew){ // if container is new is checked in method below
            if(!canCreateContainers(authentication, targetDomainObject.getContainers())){
                return false;
            }
        }

        return canSave;
    }

    private boolean canDelete(Authentication authentication, Containable targetDomainObject,
            String permissionTarget) {
        return (hasRole(authentication, "delete_created_" + permissionTarget)
                && isCreator(authentication, targetDomainObject))
                || hasRole(authentication, "delete_" + permissionTarget);
    }

    private boolean canCreateCompound(Authentication authentication,
            ChemicalCompound compound) {

        boolean isNewCompound = compound.getId() == null;
        String compoundClassName = compound.getClass().getSimpleName();
        return (hasRole(authentication, "create_" + compoundClassName) && isNewCompound);
    }

    private boolean canCreateContainers(Authentication authentication,
            Set<ChemicalCompoundContainer> containers) {

        boolean canCreateContainers = true;
        Iterator<ChemicalCompoundContainer> containerIterator = containers.iterator();
        while (canCreateContainers && containerIterator.hasNext()) {
            ChemicalCompoundContainer container = containerIterator.next();
            boolean isNewContainer = container.getId() == null;
            String containerClassName = container.getClass().getSimpleName();
            canCreateContainers = canCreateContainers && (hasRole(authentication, "create_" + containerClassName) && isNewContainer);
        }
        return canCreateContainers;
    }

    private boolean hasRole(Authentication authentication, String role) {
        Collection<? extends GrantedAuthority> authorities = roleHierarchy.getReachableGrantedAuthorities(authentication.getAuthorities());
        for (GrantedAuthority authoritiy : authorities) {
            if (authoritiy.getAuthority().equals(role)) {
                return true;
            }
        }
        return false;
    }

    private boolean isCreator(Authentication authentication, Object targetDomainObject) {

        String userName = getLogin(authentication);
        return ((Containable) targetDomainObject).getCreatedBy().equals(userName);
    }

    private String getLogin(Authentication authentication) {
        Object authenticationUserDetails = authentication.getPrincipal();
        return ((UserDetails) authenticationUserDetails).getUsername();
    }

    private boolean isAuthenticated(Authentication authentication) {
        return authentication != null && authentication.getPrincipal() instanceof UserDetails;
    }
}
