/*
 * Copyright (C) 2012 Joos Kiener
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.google.common.base.Preconditions;
import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.path.PathBuilder;
import java.beans.Introspector;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.PersistenceUnitUtil;
import lombok.AccessLevel;
import lombok.Getter;
import org.bitbucket.kienerj.moleculedatabaseframework.MoleculeDatabaseFrameworkException;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalCompoundFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.chemistry.ChemicalStructureFactory;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalStructure;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Role;
import org.bitbucket.kienerj.moleculedatabaseframework.io.EntityImportResult;
import org.bitbucket.kienerj.moleculedatabaseframework.io.IndigoMoleculeRenderer;
import org.bitbucket.kienerj.moleculedatabaseframework.io.RawSdfReader;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfCompoundCreator;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfCompoundExporter;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfRecord;
import org.bitbucket.kienerj.moleculedatabaseframework.repository.ChemicalCompoundRepository;
import org.bitbucket.kienerj.moleculedatabaseframework.repository.RoleRepository;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.SimilarityType;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.StructureSearchType;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * <p> Service class for data access of any type of
 * <code>ChemicalCompound</code>.</p> <p> To create a service for your
 * <code>ChemicalCompound</code> implementations, your need to extend
 * <code>ChemicalCompoundService</code> interface and create an implementation
 * extending this abstract class and implementing your service interface. </p>
 * <p> Example:<br>
 * <pre>
 * <code>
 *      public interface RegistrationCompoundService extends ChemicalCompoundService<RegistrationCompound>{
 *
 *          // put your custom methods here (don't forget Spring security annotations!
 *
 *          @PreAuthorize("hasRole('read_' + #root.this.getCompoundClassSimpleName())")
 *           List<RegistrationCompound> findByRegNumberStartingWith(String regNumber);
 *      }
 *
 *      public class RegistrationCompoundServiceImpl extends ChemicalCompoundServiceImpl<RegistrationCompound>
 *               implements RegistrationCompoundService {
 *          // implement abstract methods
 *      }
 * </code>
 * </pre> </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Service
@Transactional(readOnly = true)
public abstract class ChemicalCompoundServiceImpl<T extends ChemicalCompound<? extends Containable>>
        implements ChemicalCompoundService<T> {

    private static final XLogger logger = XLoggerFactory.getXLogger("ChemicalCompoundService");
    private static final boolean LOAD_COMPOSITIONS = true;
    private static final boolean LOAD_CONTAINABLES = true;
    @Getter(AccessLevel.PROTECTED)
    @PersistenceContext
    private EntityManager entityManager;
    @Autowired
    private ChemicalStructureService chemicalStructureService;
    @Getter
    @Autowired
    private ChemicalStructureFactory chemicalStructureFactory;
    @Autowired
    private RoleRepository roleRepository;
    @Getter
    private Class<T> entityClass;
    @Getter
    private PathBuilder<T> pathBuilder;
    @Getter
    private final String compoundClassSimpleName;
    // required for Spring Security SPEL expressions
    @Getter
    private final String readRole;

    public ChemicalCompoundServiceImpl(Class<T> compoundClass) {
        Preconditions.checkNotNull(compoundClass);
        this.entityClass = compoundClass;
        this.compoundClassSimpleName = compoundClass.getSimpleName();
        this.readRole = "read_" + compoundClassSimpleName;
        String path = Introspector.decapitalize(compoundClass.getSimpleName());
        pathBuilder = new PathBuilder<>(compoundClass, path);
    }

    /**
     *
     * @return Returns the spring data repository used by this service.
     */
    protected abstract ChemicalCompoundRepository<T> getRepository();

    /**
     * <p> This method is called in the save() method before the entity
     * <code>compound</code> is actually passed down to the JPA provider for
     * actual persisting (or merging). In case the entity requires some
     * operations or checks to be done before saving it, do it in this method.
     * </p> <p> Override this method if you need to. Per default it does nothing
     * at all. </p> <p> An example is setting a non-nullable value to a database
     * generated value if the value was not supplied by the user. </p>
     *
     * @param compound
     */
    protected void preSave(T compound) {
    }

    /**
     * <p> Called in save() method immediately after the entity was saved. </p>
     * <p> This can be used as example to force a flush after every save. Per
     * default flush only happens at end of transaction. </p>
     */
    protected void postSave(T compound) {
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public abstract T getExistingCompound(SdfRecord sdfRecord,
            Map<String, String> mappedCompoundProperties);

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<ChemicalCompoundComposition> getCompositions(T compound) {
        logger.entry(compound);
        Preconditions.checkNotNull(compound);
        // can't just call compound.getCompositions() as there is no hibernate session
        // opened. -> LazyInitializationException is possible
        // Also selecting all compostions:
        // select comp from ChemicalCompoundComposition comp where comp.pk.compound = ?1
        // does not work because comp.pk.compound is always a proxy due to polymorphic relationship
        // eg. polymorphic , lazy ManyToOne is problematic.
        compound = getById(compound.getId(), true, false);
        List<ChemicalCompoundComposition> result = new ArrayList<>(compound.getCompositions());
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<ChemicalCompoundComposition> getCompositions(Long compoundId) {
        logger.entry(compoundId);
        Preconditions.checkNotNull(compoundId);
        T compound = getById(compoundId, true, false);
        List<ChemicalCompoundComposition> result = new ArrayList<>(compound.getCompositions());
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getById(Long compoundId, boolean loadCompositions, boolean loadContainables) {
        logger.entry(compoundId, loadCompositions, loadContainables);
        Preconditions.checkNotNull(compoundId);
        T compound = getRepository().findOne(compoundId);
        loadLazyCollections(compound, loadCompositions, loadContainables);
        logger.exit(compound);
        return compound;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getById(Long compoundId) {
        logger.entry(compoundId);
        T compound = getById(compoundId, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(compound);
        return compound;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> getById(List<Long> compoundIds, boolean loadCompositions, boolean loadContainables) {
        logger.entry(compoundIds, loadCompositions, loadContainables);
        Preconditions.checkNotNull(compoundIds);
        List<T> compounds = toList(getRepository().findAll(compoundIds), loadCompositions, loadContainables);
        logger.exit(compounds);
        return compounds;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> getById(List<Long> ids) {
        logger.entry(ids);
        List<T> compounds = getById(ids, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(compounds);
        return compounds;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getByCas(String cas, boolean loadCompositions, boolean loadContainables) {
        logger.entry(cas, loadCompositions, loadContainables);
        Preconditions.checkNotNull(cas);
        T compound = getRepository().findByCas(cas);
        loadLazyCollections(compound, loadCompositions, loadContainables);
        logger.exit(compound);
        return compound;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getByCas(String cas) {
        logger.entry(cas);
        T compound = getByCas(cas, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(compound);
        return compound;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> getAllIds(Predicate predicate) {
        logger.entry(predicate);
        List<Long> ids = getRepository().findAll(predicate, pathBuilder);
        logger.exit(ids);
        return ids;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T findOne(Predicate predicate, boolean loadCompositions, boolean loadContainables) {
        logger.entry(predicate, loadCompositions, loadContainables);
        Preconditions.checkNotNull(predicate);
        T compound = getRepository().findOne(predicate);
        loadLazyCollections(compound, loadCompositions, loadContainables);
        logger.exit(compound);
        return compound;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T findOne(Predicate predicate) {
        logger.entry(predicate);
        T compound = findOne(predicate, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(compound);
        return compound;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate, boolean loadCompositions, boolean loadContainables) {
        logger.entry(predicate, loadCompositions, loadContainables);
        Preconditions.checkNotNull(predicate);
        List<T> result = toList(getRepository().findAll(predicate), loadCompositions, loadContainables);
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate) {
        logger.entry(predicate);
        List<T> result = findAll(predicate, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate,
            boolean loadCompositions, boolean loadContainables,
            OrderSpecifier<?>... orders) {
        logger.entry(predicate, loadCompositions, loadContainables, orders);
        // if predicate is null, all records would be returned
        // potetial risk for OutOfMemoryError
        // of course there can be predicates with same issue (compoundName like %)
        Preconditions.checkNotNull(predicate);
        List<T> result = toList(getRepository().findAll(predicate, orders), loadCompositions, loadContainables);
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final List<T> findAll(Predicate predicate,
            OrderSpecifier<?>... orders) {
        logger.entry(predicate, orders);
        List<T> result = findAll(predicate, LOAD_COMPOSITIONS, LOAD_CONTAINABLES, orders);
        logger.exit(result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findAll(Predicate predicate, Pageable pageable,
            boolean loadCompositions, boolean loadContainables) {
        logger.entry(predicate, pageable, loadCompositions, loadContainables);
        Preconditions.checkNotNull(pageable);
        // predicate allowed to be null due to paging
        Page<T> page = getRepository().findAll(predicate, pageable);
        loadLazyCollections(page, loadCompositions, loadContainables);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findAll(Predicate predicate, Pageable pageable) {
        logger.entry(predicate, pageable);
        Page<T> page = findAll(predicate, pageable, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable) {
        logger.entry(structureData, searchType, pageable);
        Page<T> page = findByChemicalStructure(structureData, searchType, pageable,
                null, "", LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate) {
        logger.entry(structureData, searchType, pageable, predicate);
        Page<T> page = findByChemicalStructure(structureData, searchType, pageable,
                predicate, "", LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate,
            String searchOptions) {
        logger.entry(structureData, searchType, pageable, predicate, searchOptions);
        Page<T> page = findByChemicalStructure(structureData, searchType, pageable,
                predicate, searchOptions, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate,
            String searchOptions, boolean loadCompositions, boolean loadContainables) {

        logger.entry(structureData, searchType, pageable, predicate, searchOptions,
                loadCompositions, loadContainables);
        // predicate may be null hence don't check
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);
        Preconditions.checkNotNull(pageable);
        Page<T> page = getRepository().findByChemicalStructure(structureData, searchType,
                pageable, predicate, searchOptions, pathBuilder);
        loadLazyCollections(page, loadCompositions, loadContainables);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable) {

        logger.entry(structureData, similarityType, lowerBound, upperBound, pageable);
        Page<T> page = this.findBySimilarStructure(structureData, similarityType,
                lowerBound, upperBound, pageable, null, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType) {

        return findByChemicalStructure(structureData, searchType, null, "");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate) {

        return findByChemicalStructure(structureData, searchType, predicate, "");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            OrderSpecifier<?>... orders) {

        return findByChemicalStructure(structureData, searchType, predicate, "", orders);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            String searchOptions, OrderSpecifier<?>... orders) {

        logger.entry(structureData, searchType);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(searchType);

        List<Long> ids = getRepository().findByChemicalStructure(structureData,
                searchType, predicate, searchOptions, pathBuilder, orders);
        logger.exit(ids);
        return ids;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate) {

        logger.entry(structureData, similarityType, lowerBound, upperBound, pageable, predicate);
        Page<T> page = this.findBySimilarStructure(structureData, similarityType,
                lowerBound, upperBound, pageable, predicate, LOAD_COMPOSITIONS, LOAD_CONTAINABLES);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate,
            boolean loadCompositions, boolean loadContainables) {
        logger.entry(structureData, similarityType, lowerBound, upperBound,
                pageable, predicate, loadCompositions, loadContainables);
        Preconditions.checkNotNull(structureData);
        Preconditions.checkNotNull(similarityType);
        Preconditions.checkNotNull(lowerBound); // upperBound can be null (=default of 1.0)
        Preconditions.checkNotNull(pageable);
        Page<T> page = getRepository().findBySimilarStructure(structureData, similarityType,
                lowerBound, upperBound, pageable, predicate, pathBuilder);
        loadLazyCollections(page, loadCompositions, loadContainables);
        logger.exit(page);
        return page;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final long count(Predicate predicate) {
        logger.entry(predicate);
        long count = getRepository().count(predicate);
        logger.exit(count);
        return count;
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @CacheEvict(value = {"containerStructureQuery", "containableStructureQuery", "compoundStructureQuery"}, allEntries = true)
    @Override
    public final T save(T compound) {
        logger.entry(compound);
        Preconditions.checkNotNull(compound);
        preSave(compound);
        if (compound.getCompositions() != null) {
            // only update compostions if the lazy collection was actually loaded
            // if no check is done an exception could be thrown if collection
            // has never been initialized.
            //Note: entityManager.merge(compound) won't work as we do not want to
            // update any ChemicalStructures automatically!!!
            PersistenceUnitUtil unitUtil = entityManager.getEntityManagerFactory().getPersistenceUnitUtil();
            if (unitUtil.isLoaded(compound, "compositions")) {
                for (ChemicalCompoundComposition composition : compound.getCompositions()) {
                    ChemicalStructure structure = composition.getChemicalStructure();
                    ChemicalStructure result = chemicalStructureService.save(structure);
                    composition.setChemicalStructure(result);
                }
            }
        }
        if (compound.getReadRole() == null) {
            Role role = roleRepository.findByRoleName("read_" + getCompoundClassSimpleName());
            if (role == null) {
                role = new Role("read_" + getCompoundClassSimpleName());
                role = roleRepository.save(role);
            }
            compound.setReadRole(role);
        }

        compound = getRepository().save(compound);
        postSave(compound);
        logger.exit(compound);
        return compound;
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final EntityImportResult importSDF(final Reader reader,
            Map<String, String> mappedProperties,
            final boolean abortOnChemistryError) {
        return importSDF(reader, mappedProperties, null, null, abortOnChemistryError);
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final EntityImportResult importSDF(final Reader reader, final boolean abortOnChemistryError) {
        return importSDF(reader, null, null, null, abortOnChemistryError);
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final EntityImportResult importSDF(final Reader reader,
            final Map<String, String> mappedProperties,
            final Map<String, Map<String, Double>> percentages,
            final String recordIdentifier, final boolean abortOnChemistryError) {

        logger.entry(reader, mappedProperties, percentages, recordIdentifier, abortOnChemistryError);
        Preconditions.checkNotNull(reader);
        final RawSdfReader sdfReader = new RawSdfReader(reader);

        final ChemicalCompoundFactory<T> compoundFactory =
                new ChemicalCompoundFactory<>(chemicalStructureFactory, entityClass);

        final EntityImportResult importResult = new EntityImportResult();
        int numWorkerThreads = 1; // more do not increase performance
        int maxQueueSize = 1000;

        final ExecutorService taskGenerator = Executors.newFixedThreadPool(1);
        final ExecutorService taskExecutor = Executors.newFixedThreadPool(numWorkerThreads);
        final LinkedBlockingQueue<Future<T>> futures = new LinkedBlockingQueue<>(maxQueueSize);

        Runnable generator = new Runnable() {
            @Override
            public void run() {

                while (sdfReader.hasNext()) {

                    SdfRecord sdfRecord = sdfReader.next();

                    // if compound already exists, use that compound
                    // Note that DataIntegrityViolationException (constraint violation)
                    // is not directly thrown but only when the etity is actually
                    // saved to database possibly after this methods was left.
                    // hence we first check if an entity already exists instead of
                    // handling the exception.
                    //Note: getExistingCompound runs in a different thread and hence
                    // a different transaction!!!
                    T existingCompound = getExistingCompound(sdfRecord, mappedProperties);
                    if (existingCompound != null) {
                        importResult.addAlreadyPresentEntity(existingCompound.getId());
                        logger.debug("Found existing compound {}.", existingCompound.getCompoundName());
                        continue;
                    }

                    Map<String, Double> currentPercentages = null;
                    if (recordIdentifier != null && percentages != null) {
                        String identifier = sdfRecord.getProperties().get(recordIdentifier);
                        currentPercentages = percentages.get(identifier);
                        logger.debug("Found percentages for compound {}.", identifier);
                    }

                    SdfCompoundCreator<T> sdfCompoundCreator =
                            new SdfCompoundCreator<>(sdfRecord, compoundFactory,
                            mappedProperties, currentPercentages, importResult,
                            abortOnChemistryError);
                    Future<T> compoundFuture =
                            taskExecutor.submit(sdfCompoundCreator);
                    try {
                        futures.put(compoundFuture);
                    } catch (InterruptedException ex) {
                        Thread.currentThread().interrupt();
                        throw new MoleculeDatabaseFrameworkException(ex);
                    }
                }
                taskExecutor.shutdown();
            }
        };
        try {
            Future<?> generatorResult = taskGenerator.submit(generator);

            int counter = 0;
            while (!generatorResult.isDone()
                    || !futures.isEmpty()) {
                Future<T> compoundFuture = futures.take();
                T compound = compoundFuture.get();
                compound = save(compound);
                importResult.addImportedEntity(compound.getId());
                counter++;
                // Set hibernate.jdbc.batch_size to 100 too
                // massive performance increase with large inserts
                // at least 10 times faster
                if (counter >= 100) {
                    counter = 0;
                    entityManager.flush();
                    entityManager.clear();
                    logger.debug("Batch Inserting: entityManager.flush() and entityManager.clear()");
                }
            }
            // if not called all exceptions thrown in task generator
            // will be "lost" and method would complete successfully.
            generatorResult.get();
            logger.exit(importResult);
            return importResult;
        } catch (InterruptedException ex) {
            Thread.currentThread().interrupt();
            throw new MoleculeDatabaseFrameworkException(ex);
        } catch (ExecutionException ex) {
            throw new MoleculeDatabaseFrameworkException(ex);
        } finally {
            taskExecutor.shutdownNow();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void exportSDF(Collection<Long> compoundIds, Writer writer) {

        logger.entry(compoundIds, writer);
        Preconditions.checkNotNull(writer);
        Preconditions.checkNotNull(compoundIds);

        SdfCompoundExporter<T> exporter = new SdfCompoundExporter<>(this);

        exporter.exportCompoundsToSdf(compoundIds, writer, null);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void exportSDF(Collection<Long> compoundIds, Writer writer,
            Set<String> propertiesToExport) {

        logger.entry(compoundIds, writer, propertiesToExport);
        Preconditions.checkNotNull(writer);
        Preconditions.checkNotNull(compoundIds);
        Preconditions.checkNotNull(propertiesToExport);

        SdfCompoundExporter<T> exporter = new SdfCompoundExporter<>(this);

        exporter.exportCompoundsToSdf(compoundIds, writer, propertiesToExport);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void exportSDF(Collection<Long> compoundIds, Writer writer,
            Set<String> propertiesToExport, String dateFormat) {

        logger.entry(compoundIds, writer, propertiesToExport, dateFormat);
        Preconditions.checkNotNull(writer);
        Preconditions.checkNotNull(compoundIds);
        Preconditions.checkNotNull(propertiesToExport);
        Preconditions.checkNotNull(dateFormat);

        SdfCompoundExporter<T> exporter = new SdfCompoundExporter<>(this);

        exporter.exportCompoundsToSdf(compoundIds, writer, propertiesToExport, dateFormat);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void renderChemicalCompound(Long compoundId, OutputStream out,
            int width, int height) {

        this.renderChemicalCompound(compoundId, out, width, height, true);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void renderChemicalCompound(Long compoundId, OutputStream out,
            int width, int height, boolean fixedWidth) {

        IndigoMoleculeRenderer renderer = new IndigoMoleculeRenderer(width, height);
        renderer.renderChemicalCompound(getById(compoundId), out, !fixedWidth);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void renderChemicalCompound(Long compoundId, OutputStream out,
            int width, int height, boolean fixedWidth, String highlightedSubstructure) {

        IndigoMoleculeRenderer renderer = new IndigoMoleculeRenderer(width, height);
        renderer.renderChemicalCompound(getById(compoundId), out, highlightedSubstructure, !fixedWidth);
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final void delete(T compound) {
        logger.entry(compound);
        Preconditions.checkNotNull(compound);
        getRepository().delete(compound);
        logger.exit();
    }

    /**
     * {@inheritDoc}
     */
    @Transactional(readOnly = false)
    @Override
    public final void deleteById(Long compoundId) {
        logger.entry(compoundId);
        Preconditions.checkNotNull(compoundId);
        getRepository().delete(compoundId);
        logger.exit();
    }

    private List<T> toList(Iterable<T> iterable, boolean loadCompositions, boolean loadContainables) {
        if (iterable == null) {
            return null;
        }
        List<T> compounds = new ArrayList<>();
        for (T compound : iterable) {
            compounds.add(compound);
            loadLazyCollections(compound, loadCompositions, loadContainables);
        }
        return compounds;
    }

    private void loadLazyCollections(Page<T> page, boolean loadCompositions, boolean loadContainables) {
        if (page != null && (loadCompositions || loadContainables)) {
            for (T compound : page) {
                loadLazyCollections(compound, loadCompositions, loadContainables);
            }
        }
    }

    private void loadLazyCollections(T compound, boolean loadCompositions, boolean loadContainables) {
        if (compound != null) {
            if (loadCompositions) {
                for (ChemicalCompoundComposition composition : compound.getCompositions()) {
                    composition.getChemicalStructure().getId();
                }
            }
            if (loadContainables) {
                for (Containable containable : compound.getContainables()) {
                    containable.getId();
                }
            }
        }
    }
}
