/*
 *  Copyright (C) 2012 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.mysema.query.types.Predicate;
import java.io.OutputStream;
import java.util.List;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompoundComposition;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalStructure;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.transaction.annotation.Transactional;

/**
 * <p>
 * Service interface for data access of <code>ChemicalStructure</code>.
 * </p>
 * <p>
 * All service methods are transactional and annotated with Spring Security. Security
 * must be explictly enabled in configuration.
 * </p>
 * <p>
 * For saving <code>ChemicalStructures</code> always use this service and never
 * the repository directly!
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Transactional(readOnly=true)
public interface ChemicalStructureService<T extends ChemicalStructure>
        extends Service<T> {

    /**
     * {@inheritDoc}
     */
    @PreAuthorize("hasRole('read_ChemicalStructure')")
    @Override
    T getById(Long id);

    /**
     * Get ChemicalStructure by its unique structure key
     */
    @Cacheable("chemicalStructure")
    @PreAuthorize("hasRole('read_ChemicalStructure')")
    T getByStructureKey(String structureKey);

    /**
     * <p>
     * Creates or updates <code>structure</code>. After calling this method it
     * is crucial to continue work with the returned structure (which can be a
     * different instance). So always call this method like below:<pre>
     *
     * <code>
     *
     * structure = chemicalStructureService.save(structure);
     *
     * </code>
     *
     * (google for JPA merge() vs. persist())
     * </pre>
     * </p>
     * <p>
     * Note that <code>ChemicalStructures</code> are immutable and the passed
     * in <code>structure</code> is never changed. The method will either create
     * a new <code>ChemicalStructure</code> or return an existing one that is
     * equal to <code>structure</code>.
     * </p>
     * @param structure
     * @return the passed in ChemicalStructure fully updated.
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasRole('save_ChemicalStructure')")
    @Override
    T save(T structure);


    /**
     * <p>
     * Updates an existing <code>ChemicalStructure</code>. Use with care. This
     * will affect every single <code>ChemicalCompoundComposition</code> and hence
     * <code>ChemicalCompound</code> that uses <code>structure</code>.
     * </p>
     * <p>
     * This method should only ever be called if you want to update
     * <code>structure</code> for all <code>ChemicalCompoundCompositions</code>
     * it appears in. If you want to update a single
     * <code>ChemicalCompoundComposition</code>, use <code>save(structure)</code> instead.
     * </p>
     * <p>
     * After calling this method, always use the return value and never the original one:
     * <code>
     *
     * structure = chemicalStructureService.updateExistingStructure(structure);
     *
     * </code>
     * </p>
     * @param structure the structure to update
     * @return the updated structure.
     */
    @Transactional(readOnly = false)
    // evict structure from 2nd level hibernate cache
    @CacheEvict(value = {"chemicalStructure", "containerStructureQuery",
        "compoundStructureQuery"}, allEntries = true)
    @PreAuthorize("hasRole('update_ChemicalStructure')")
    T updateExistingStructure(T structure);

    /**
     * Counts number of structures matching the given predicate
     *
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole('read_ChemicalStructure')")
    long count(Predicate predicate);

    /**
     * <p>
     * Returns a List of <code>ChemicalCompounds</code>
     * <code>structure</code> occurs in but not the percentage. If percentage is
     * required, call <code>getCompositions(structure)</code> instead.
     * </p>
     * <p>
     * Does not load lazy collections of returned compounds. To load them you need
     * to call
     * </p>
     * <pre>
     * <code>
     *      compoundService.getCompositions(compound);
     *      compoundService.getContainables(compound);
     * </code>
     * </pre>
     * or you LazyInitializationException is thrown outside of a entitiyManger session.
     * @param structure
     * @return
     */
    @PreAuthorize("hasRole('read_ChemicalStructure')")
    List<ChemicalCompound> getOccurences(ChemicalStructure structure);

    /**
     * Returns a List of ChemicalCompoundCompostion <code>structure</code> is part of.
     *
     *
     * @param structure
     * @return
     */
    @PreAuthorize("hasRole('read_ChemicalStructure')")
    List<ChemicalCompoundComposition> getCompositions(ChemicalStructure structure);


    /**
     * <p>
     * Returns a List of <code>ChemicalCompounds</code>
     * <code>CHemicalStructure</code> with id <code>structureId</code> occurs in
     * but not the percentage. If percentage is
     * required, call <code>getCompositions(structureId)</code> instead.
     * </p>
     * <p>
     * Does not load lazy collections of returned compounds. To load them you need
     * to call
     * </p>
     * <pre>
     * <code>
     *      compoundService.getCompositions(compound);
     *      compoundService.getContainables(compound);
     * </code>
     * </pre>
     * or you LazyInitializationException is thrown outside of a entitiyManger session.
     * @param structureId
     * @return
     */
    @PreAuthorize("hasRole('read_ChemicalStructure')")
    List<ChemicalCompound> getOccurences(Long structureId);

    /**
     * <p>
     * Returns a List of <code>ChemicalCompoundCompostions</code> the
     * <code>CHemicalStructure</code> with id <code>structureId</code> is part of.
     * </p>
     *
     * @param structureId
     * @return
     */
    @PreAuthorize("hasRole('read_ChemicalStructure')")
    List<ChemicalCompoundComposition> getCompositions(Long structureId);


    /**
     * <p>
     * Checks if <code>structure</code> violates any unique constraints, in case
     * of <code>ChemicalStructure</code> this is always only the structure key.
     * </p>
     * <p>
     * If <code>returnValue.isUnique() == true</code> saving this structure
     * will create a new <code>ChemicalStructure</code>. In case of <code>false</code>
     * the existing <code>ChemicalStructure</code> will be reused.
     * </p>
     * <p>
     * The returned value contains a mapping of all fields an their values
     * that violate uniqueness, here always only the structure key.
     * </p>
     *
     * @param compound
     */
    @Override
    UniquenesscheckResult checkUniqueness(T structure);


    /**
     * <p>Renders a png image of the desired <code>ChemicalStructure</code> to
     * <code>out</code>.</p>
     *
     * @param structureId the compounds unique database id
     * @param out the stream to which to ouput the image
     * @param width the width of the image but depends on fixedWidth
     * @param height the total height of the image.
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    void renderChemicalStructure(Long structureId, OutputStream out,
            int width, int height);

    /**
     * <p>Renders a png image of the desired <code>ChemicalStructure</code> to
     * <code>out</code>.</p>
     *
     * <p>The part matching <code>highlightedSubstructure</code>
     * is highlighted.</p>
     *
     * @param structureId the compounds unique database id
     * @param out the stream to which to ouput the image
     * @param width the width of the image but depends on fixedWidth
     * @param height the total height of the image.
     * @param highlightedSubstructure the part of the structure to highlight
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    void renderChemicalStructure(Long structureId, OutputStream out,
            int width, int height, String highlightedSubstructure);


    /**
     *
     * @param structureData SMILES or molfile
     * @return
     */
    @PreAuthorize("isAuthenticated()")
    String getStandardInChi(String structureData);

    /**
     *
     * @param structureData SMILES or molfile
     * @return
     */
    @PreAuthorize("isAuthenticated()")
    String getStandardInChiKey(String structureData);


    /**
     * Gets an InChi. If options is null or empty string, the result will
     * be the same as <code>getStandardInChi(structureData)</code>
     *
     * See <a href="http://www.inchi-trust.org/fileadmin/user_upload/html/inchifaq/inchi-faq.html#15.17">Information on InChi options</a>.
     *
     * @param structureData SMILES or molfile
     * @param options options for generating InChi
     * @return
     */
    @PreAuthorize("isAuthenticated()")
    String getInChi(String structureData, String options);

    /**
     * Gets an InChi key. If options is null or empty string, the result will
     * be the same as <code>getStandardInChiKey(structureData)</code>
     *
     * See <a href="http://www.inchi-trust.org/fileadmin/user_upload/html/inchifaq/inchi-faq.html#15.17">Information on InChi options</a>.
     *
     * @param structureData SMILES or molfile
     * @param options options for generating InChi
     * @return
     */
    @PreAuthorize("isAuthenticated()")
    String getInChiKey(String structureData, String options);
}
