/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.service;

import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import java.io.Reader;
import java.io.Writer;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.ChemicalCompound;
import org.bitbucket.kienerj.moleculedatabaseframework.entity.Containable;
import org.bitbucket.kienerj.moleculedatabaseframework.io.EntityImportResult;
import org.bitbucket.kienerj.moleculedatabaseframework.io.SdfRecord;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.SimilarityType;
import org.bitbucket.kienerj.moleculedatabaseframework.structuresearch.StructureSearchType;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.transaction.annotation.Transactional;

/**
 * <p>
 * Service interface for data access of any type of <code>Containable</code>.
 * </p>
 * <p>
 * All service methods are transactional and annotated with Spring Security. Security
 * must be explictly enabled in configuration.
 * </p>
 * <p>
 * To create a service for your <code>Containable</code> implementations,
 * your need to extend this interface and create an implementation extending
 * <code>ContainableServiceImpl</code> and implementing your service interface.
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
@Transactional(readOnly=true)
public interface ContainableService<T extends Containable> extends Service<T> {

    /**
     * Gets a containable by its internal id
     *
     * @param id
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    @Override
    T getById(final Long id);

    /**
     * Gets all containables with an id that is contained in
     * <code>ids</code>.
     *
     * @param ids
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> getById(final List<Long> ids);

    /**
     * Get Containables by contained <code>ChemicalCompound</code>
     *
     * @param compound
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> getByChemicalCompound(ChemicalCompound<T> compound);


    /**
     * <p>
     * Get Containables by a <code>ChemicalCompounds</code> id.
     * </p>
     * <p>
     * This can be useful in AJAX web applications and master detail views
     * if you don't want to load all Containables directly.
     * </p>
     * @param compound
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> getByChemicalCompoundId(Long compoundId);

    /**
     * <p>
     * Returns an unsorted List of all containables id matching the predicate.
     * </p>     *
     * <p>
     * <code>predicate</code> must not be null or an NullPointertException is thrown.
     * </p>     *
     * <p>
     * If security is enabled, requires read role for this Containable
     * implementation (eg. read_TestContainable)
     * </p>
     *
     * @param predicate the filter (where clause) for determining which containables to return. must not be null.
     * @return a <code>List</code> of all containables ids matching the given <code>predicate</code>
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> getAllIds(Predicate predicate);

    /**
     * Returns 1 containable matching the given predicate.
     *
     * @param predicate
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    T findOne(Predicate predicate);

    /**
     * <p>
     * Returns an unsorted List of all containables matching the predicate.
     * </p>
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     * </p>
     *
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate);

    /**
     * Returns a sorted List of all containables matching the predicate.
     *
     * <p>
     * Use with care as this will load all entities matching the predicate
     * into memory.
     * </p>
     *
     * @param predicate
     * @param orders
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<T> findAll(Predicate predicate, OrderSpecifier<?>... orders);

    /**
     * Returns a Page of containables matching the given predicate.
     *
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findAll(Predicate predicate, Pageable pageable);

    /**
     * <p>
     * Returns a Page of containers that contain at least 1 chemical structure
     * matching the chemical structure search arguments.
     * </p>
     * <p>
     * Example of Pageable:
     * </p>
     * <pre>
     * <code>
     * PageRequest pageable = new PageRequest(
     *           offset, limit, new Sort(
     *           new Order(Direction.ASC, "containerName")));
     * </code>
     * </pre>
     * <p>
     * The name given to Order ("containerName") must equal a field name
     * of ChemicalCompoundContainer class.
     * </p>
     *
     * @param structureData
     * @param searchType
     * @param pageable
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable);

    /**
     * <p>
     * Returns a Page of containers that contain at least 1 chemical structure
     * matching the chemical structure search arguments and the predicate.
     * </p>
     * <p>
     * Example of Pageable:
     * </p>
     * <pre>
     * <code>
     * PageRequest pageable = new PageRequest(
     *           offset, limit, new Sort(
     *           new Order(Direction.ASC, "containerName")));
     * </code>
     * </pre>
     * <p>
     * The name given to Order ("containerName") must equal a field name
     * of ChemicalCompoundContainer class.
     * </p>
     *
     * <p>A Predicate is a set of Expressions combined with boolean operator.</p>
     * <p>A Predicate is created by using the querydsl generated query classes.</p>
     *
     * <p>Example:</p>
     * <pre>
     * <code>
     *
     * QBasicCompoundContainer basicContainer =
                QBasicCompoundContainer.basicCompoundContainer;
        QChemicalCompound compound = QChemicalCompound.chemicalCompound;
        QChemicalCompoundComposition composition =
                QChemicalCompoundComposition.chemicalCompoundComposition;
     *
     * Predicate predicate =
     *     container.containerName.like("Methyl%").and(composition.percentage.gt(50));
     *
     * </code>
     * </pre>
     * @param structureData
     * @param searchType
     * @param pageable
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate);

    /**
     * <p>
     * See
     * {@link #findByChemicalStructure(String, StructureSearchType,
     * Pageable, Predicate) findByChemicalStructure} for detailed information.
     *</p>
     * <p>
     * For possible values for
     * <code>searchOptions</code> see the
     * <a href="http://ggasoftware.com/opensource/bingo/user-manual-postgres#substructure-search">Bingo PostgreSQL cartridge documentation</a>.
     * </p>
     *
     * @param structureData
     * @param searchType
     * @param pageable
     * @param predicate
     * @param searchOptions
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate,
            String searchOptions);

    /**
     * <p>
     * See
     * {@link #findByChemicalStructure(String, StructureSearchType,
     * Pageable, Predicate) findByChemicalStructure} for detailed information.
     * </p>
     * <p>
     * For possible values for
     * <code>searchOptions</code> see the
     * <a href="http://ggasoftware.com/opensource/bingo/user-manual-postgres#substructure-search">Bingo PostgreSQL cartridge documentation</a>.
     * </p>
     * <p>
     * if loadCompositions is true, all the <code>ChemicalCompoundCompositions</code>
     * of the <code>ChemicalCompound</code> associated with a <code>Containable</code>
     * are loaded from database.
     * </p>
     *
     * @param structureData
     * @param searchType
     * @param pageable
     * @param predicate
     * @param searchOptions
     *
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findByChemicalStructure(String structureData,
            StructureSearchType searchType,
            Pageable pageable, Predicate predicate,
            String searchOptions, boolean loadCompositions);

/**
     * <p>Returns a list of ids for containables matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType);

    /**
     * <p>Returns a list of ids for containables matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate);

    /**
     * <p>Returns a list of ids for containables matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            OrderSpecifier<?>... orders);

    /**
     * <p>Returns a list of ids for containables matching the search criteria.</p>
     *
     * @param structureData
     * @param searchType
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    List<Long> findByChemicalStructure(String structureData,
            StructureSearchType searchType, Predicate predicate,
            String searchOptions, OrderSpecifier<?>... orders);


    /**
     * <p>
     * Search for chemical structures similar to <code>structureData</code>
     * </p>
     * <p>lowerBound and upperBound take values between 0 to 1 (including).</p>
     *
     * @param structureData
     * @param similarityType the algorithm used for determining similarity
     * @param lowerBound    minimum required similarity
     * @param upperBound    maximum allowed similarity
     * @param pageable      for sorting and paging
     * @param predicate     other criteria beside similarity
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate);


    /**
     * <p>
     * Search for chemical structures similar to <code>structureData</code>
     * </p>
     * <p>lowerBound and upperBound take values between 0 to 1 (including).</p>
     *
     * <p>
     * if loadCompositions is true, all the <code>ChemicalCompoundCompositions</code>
     * of the <code>ChemicalCompound</code> associated with a <code>Containable</code>
     * are loaded from database.
     * </p>
     *
     * @param structureData
     * @param similarityType the algorithm used for determining similarity
     * @param lowerBound    minimum required similarity
     * @param upperBound    maximum allowed similarity
     * @param pageable      for sorting and paging
     * @param predicate     other criteria beside similarity
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    Page<T> findBySimilarStructure(String structureData,
            SimilarityType similarityType,
            Double lowerBound, Double upperBound,
            Pageable pageable, Predicate predicate,
            boolean loadCompositions);


    /**
     * Counts number of containables matching given predicate.
     *
     * @param predicate
     * @return
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    long count(Predicate predicate);

    /**
     * <p>
     * Creates or updates
     * <code>containable</code>. After calling this method it is crucial to
     * continue work with the returned containable (which can be a different
     * instance). So always call this method like below:
     * </p>
     * <code>
     * containable = chemicalCompoundService.save(containable);     *
     * </code>
     * <p>
     * (google for JPA merge() vs. persist())
     * </p>
     * @param containable
     * @return
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasPermission(#containable, 'save_' + #root.this.getContainableClassSimpleName())")
    @Override
    T save(T containable);


    /**
     * * <p>
     * Imports an SD-File into the database. Each record in the sd-file is
     * considered to be a containable.
     * </p>
     * <p>
     * Properties are mapped to the according entity  (Containable, Compound)
     * by the according Maps <code>mappedCompoundProperties, mappedContainableProperties</code>.
     * </p>
     * <p>
     * The Compound in the containable can contain disconnected chemical structures.
     * The percentage of a certain structure in a compound is mapped by defining
     * a unique property <code>compoundIdentifier</code> in the sd-file. Then for each chemical
     * structure in a Compound create a mapping of the InChIkey to its
     * percentage in the compound. Add this Map to <code>percentages</code>
     * mapped by the unique property. If no mapping is found, percentages are set
     * to null. If the compound only has 1 structure, per default percentage is
     * set to 100 and no mapping is required.
     * </p>
     * <pre>
     * <code>
     *   String compoundIdentifier = "Name";
     *   Map<String,Double> singleRecordPercentage = new HashMap<>();
     *   singleRecordPercentage.put("XDTMQSROBMDMFD-UHFFFAOYSA-N", 30.0);
     *   singleRecordPercentage.put("ZSWFCLXCOIISFI-UHFFFAOYSA-N", 70.0);
     *   Map<String,Map<String,Double>> percentages = new HashMap<>();
     *   percentages.put("Mixture_1", singleRecordPercentage);
     * </code>
     * </pre>
     * <p>
     * Properties are mapped by name (case-sensitive) were the key
     * is the property in the sd-file and the value is the field name of the
     * entity.<br> <br> If compoundName is not mapped, it is set
     * to the compounds name in the sd-file. If that is null an exception will occur.
     * <br> <br> Type is resolved by the fields type.
     * Only primitives, Strings or dates can be imported. Date / Calendar values
     * must be in one of the following formats: yyyy-MM-dd HH:mm:ss<br>
     * yyyy-MM-dd<br> dd.mm.yyyy<br> mm/dd/yyy<br> where HH is the hour value
     * from 0 to 23 (see java.text.SimpleDateFormat)
     * </p>
     * <p>
     * If <code>abortOnChemistryError</code> is true, Exceptions that occur due to
     * the chemical structure being invalid remain uncaught and transaction is
     * rolled back. If false, invalid structures are ignored and import
     * continues. However exceptions due to invalid property mappings, invalid
     * values,... are always wrapped in
     * <code>MoleculeDatabaseFrameworkException</code>.
     * </p>
     *
     *
     * @param reader
     * @param mappedCompoundProperties
     * @param mappedContainableProperties
     * @param uniqueContainableProperty
     * @param percentages
     * @param compoundIdentifier
     * @param abortOnChemistryError
     * @return
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasRole('import_' + #root.this.getContainableClassSimpleName())")
    EntityImportResult importSDF(final Reader reader,
            final Map<String, String> mappedCompoundProperties,
            final Map<String, String> mappedContainableProperties,
            final Map<String, Map<String, Double>> percentages,
            final String compoundIdentifier, final boolean abortOnChemistryError);


    /**
     * <p>
     * Exports containables to an SD-File.
     * </p>
     * <p>
     * If the containable contains of multiple structures, the percentages are
     * exported as a property of name InChIkey and the value is the
     * percentage for that structure.
     * </p>
     * <p>
     *  The <code>Sets<String></code> must contain the
     *  names of all the fields you want to export for the according entity.
     *  If it is empty, only the structure will be exported, if it is null, all
     * fields will be exported.
     * </p>
     * <p>
     * For date fields default format of "yyyy-MM-dd HH:mm:ss" is used.
     * </p>
     * <p>
     * If security is enabled, requires export role for this ChemicalCompoundContainer
     * implementation (eg. export_TestCompoundContainer)
     * </p>
     *
     * @param containableIds the ids of the containables to export
     * @param writer the writer responsible for writing the output sd-file
     */
    @PreAuthorize("hasRole('export_' + #root.this.getContainableClassSimpleName())")
    void exportSDF(Collection<Long> containableIds, Writer writer,
            Set<String> containableProperties, Set<String> compoundProperties);


    /**
     * <p>
     * Exports containables to an SD-File.
     * </p>
     * <p>
     * If the containable contains of multiple structures, the percentages are
     * exported as a property of name InChIkey and the value is the
     * percentage for that structure.
     * </p>
     * <p>
     *  The <code>Sets<String></code> must contain the
     *  names of all the fields you want to export for the according entity.
     *  If it is empty, only the structure will be exported, if it is null, all
     * fields will be exported.
     * </p>
     * <p>
     * For date fields the date format defined by <code>dateFormat</code> is used.
     * See <a href="http://docs.oracle.com/javase/7/docs/api/java/text/SimpleDateFormat.html">SimpleDateFormat</a>
     * for more information.
     * </p>
     * <p>
     * If security is enabled, requires export role for this ChemicalCompoundContainer
     * implementation (eg. export_TestCompoundContainer)
     * </p>
     *
     * @param containableIds the ids of the containables to export
     * @param writer the writer responsible for writing the output sd-file
     */
    @PreAuthorize("hasRole('export_' + #root.this.getContainableClassSimpleName())")
    void exportSDF(Collection<Long> containableIds, Writer writer,
            Set<String> containableProperties, Set<String> compoundProperties,
            String dateFormat);

    /**
     * Deletes the containable from the database
     *
     * @param containable
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasPermission(#containable, 'delete_' + #root.this.getContainableClassSimpleName())")
    void delete(T containable);

    /**
     * Deletes the containable with the given id from the database
     *
     * @param containableId
     */
    @Transactional(readOnly = false)
    @PreAuthorize("hasRole('delete_' + #root.this.getContainableClassSimpleName())")
    void deleteById(Long containableId);

    /**
     * <p>
     * Checks if <code>containable</code> already exists in the database by looking
     * up all unique constrained fields.
     * </p>
     * <p>
     * If <code>returnValue.isUnique() == true</code>, the containable can be safley
     * inserted. If false, a call to save() will end in a
     * <code>DataIntegrityViolationException</code>.
     * </p>
     * <p>
     * The returned value contains a mapping of all fields an their values
     * that violate uniqueness.
     * </p>
     *
     * @param containable
     */
    @PreAuthorize("hasRole(#root.this.getReadRole())")
    @Override
    UniquenesscheckResult checkUniqueness(T containable);


    /**
     * <p>
     * Returns a containable with given input properties or null if it does not
     * exist in the database.
     * </p>
     * <p>
     * When importing containables from SD-File this method checks if an equal
     * containable already exists in the database and if yes return it. This should be
     * a "best-effort" to avoid unique constraint violation when inserting the
     * new imported entity.
     * </p>
     *
     * @param sdfRecord
     * @param mappedContainableProperties
     * @param mappedCompoundProperties
     * @return
     */
    @PreAuthorize("hasRole('import_' + #root.this.getContainableClassSimpleName())")
    T getExistingContainable(SdfRecord sdfRecord,
            Map<String, String> mappedContainableProperties,
            Map<String, String> mappedCompoundProperties);

}
