/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.sql;

import javax.persistence.OptimisticLockException;
import lombok.Getter;
import org.bitbucket.kienerj.moleculedatabaseframework.MoleculeDatabaseFrameworkException;
import org.springframework.orm.ObjectOptimisticLockingFailureException;
import org.springframework.orm.jpa.JpaOptimisticLockingFailureException;

/**
 * <p> Thrown when a
 * <code>ObjectOptimisticLockingFailureException</code> occurs. That exception
 * will always be the cause, available through
 * <code>getCause()</code>. </p> <p> The exception contains an additional field
 * <code>currentVersion</code> which is the newest available version at the time
 * the exception happens. This information should be passed to the client so the
 * conflict can be resolved. </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */
public class EntityVersionConflictException extends MoleculeDatabaseFrameworkException {

    @Getter
    private final Object currentVersion;
    @Getter
    private final Object conflictingVersion;

    public EntityVersionConflictException(String message, Object currentVersion,
            Object conflictingVersion) {
        super(message);
        this.currentVersion = currentVersion;
        this.conflictingVersion = conflictingVersion;
    }

    /**
     *
     * @param lockEx
     * @param currentVersion
     */
    public EntityVersionConflictException(
            Throwable cause,
            Object currentVersion,
            Object conflictingVersion) {
        super(cause);
        this.currentVersion = currentVersion;
        this.conflictingVersion = conflictingVersion;
    }

        /**
     * <p><code>lockEx</code> should be caused by
     * <code>OptimisticLockException</code> or else
     * <code>getConflictingVersion()
     * <code> returns null.
     *
     * @param lockEx
     * @param currentVersion
     */
    public EntityVersionConflictException(
            JpaOptimisticLockingFailureException cause,
            Object currentVersion) {
        super(cause);
        this.currentVersion = currentVersion;
        // this should always be true as cause of JpaOptimisticLockingFailureException
        // must be a OptimisticLockException
        if (getCause().getCause() instanceof OptimisticLockException) {
            this.conflictingVersion = ((OptimisticLockException) getCause().getCause()).getEntity();
        } else {
            this.conflictingVersion = null;
        }
    }

    public Class getEntityClass() {
        return getCause().getPersistentClass();
    }

    @Override
    public ObjectOptimisticLockingFailureException getCause() {
        return (ObjectOptimisticLockingFailureException) super.getCause();
    }
}
