/*
 *  Copyright (C) 2013 Joos Kiener <Joos.Kiener@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.

 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bitbucket.kienerj.moleculedatabaseframework.sql;

import java.math.BigInteger;
import java.util.Collections;
import java.util.Iterator;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;

/**
 * <p>
 * Generates a numeric string sequence with optional pre- and postfix using an
 * existing database sequence.
 * </p>
 *
 * @author Joos Kiener <Joos.Kiener@gmail.com>
 */

public class PostgreSqlSequenceGenerator {

    @PersistenceContext
    private EntityManager entityManager;
    private Iterator<BigInteger> iter;
    private final long batchSize;
    private final String sequenceName;
    private final String prefix;
    private final String prefixDelimiter;
    private final String postFix;
    private final String postFixDelimiter;
    private final String format;

    /**
     * <p>
     * Constructs a generator that will return a String with a pre- and postfix
     * and the sequence value formatted according to <code>format</code>.
     * </p>
     * <p>
     * An example format is <code>"%05d"</code>. This will left pad the sequence
     * number up to 5 zeroes, 00001.
     * </p>
     * <p>
     * See <a href="http://docs.oracle.com/javase/7/docs/api/java/util/Formatter.html#syntax">
     * String.format</a> for further information on formatting options.
     * </p>
     *
     * @param entityManager JPA entityManager used to create the query
     * @param sequenceName  the name of the sequence to get the numbers from
     * @param format how the sequence value is formatted as String
     * @param batchFetchSize how many values to fetch from database per call
     * @param prefix fixed string before sequence value
     * @param prefixDelimiter fixed string between prefix and sequence value
     * @param postFix fixed string after sequence value
     * @param postFixDelimiter fixed string between postfix and sequence value
     */
    public PostgreSqlSequenceGenerator(String sequenceName,
            String format, long batchFetchSize,
            String prefix, String prefixDelimiter,
            String postFix, String postFixDelimiter) {
        this.sequenceName = sequenceName;
        this.prefix = prefix == null ? "" : prefix;
        this.prefixDelimiter = prefixDelimiter == null ? "" : prefixDelimiter;
        this.postFix = postFix  == null ? "" : postFix;
        this.postFixDelimiter = postFixDelimiter == null ? "" : postFixDelimiter ;
        batchSize = batchFetchSize;
        iter = Collections.<BigInteger>emptyList().iterator();
        this.format = format == null ? "" : format;
    }

    /**
     * <p>
     * Constructs a generator that will return a numeric String only, formatted
     * according to <code>format</code>.
     * </p>
     * <p>
     * An example format is <code>"%05d"</code>. This will left pad the sequence
     * number up to 5 zeroes, 00001.
     * </p>
     * <p>
     * See <a href="http://docs.oracle.com/javase/7/docs/api/java/util/Formatter.html#syntax">
     * String.format</a> for further information on formatting options.
     * </p>
     * @param entityManager JPA entityManager used to create the query
     * @param sequenceName  the name of the sequence to get the numbers from
     * @param format
     * @param batchFetchSize how many values to fetch from database per call
     */
    public PostgreSqlSequenceGenerator(String sequenceName, String format,
            long batchFetchSize){
        this(sequenceName, format, batchFetchSize, "", "", "", "");
    }

    /**
     * <p>
     * Constructs a generator that will return a String with a prefix
     * and the sequence value formatted according to <code>format</code>.
     * </p>
     * <p>
     * An example format is <code>"%05d"</code>. This will left pad the sequence
     * number up to 5 zeroes, 00001.
     * </p>
     * <p>
     * See <a href="http://docs.oracle.com/javase/7/docs/api/java/util/Formatter.html#syntax">
     * String.format</a> for further information on formatting options.
     * </p>
     *
     * @param entityManager JPA entityManager used to create the query
     * @param sequenceName  the name of the sequence to get the numbers from
     * @param format how the sequence value is formatted as String
     * @param batchFetchSize how many values to fetch from database per call
     * @param prefix fixed string before sequence value
     * @param prefixDelimiter fixed string between prefix and sequence value
     */
    public PostgreSqlSequenceGenerator(String sequenceName, String prefix,
            String prefixDelimiter, String format, long batchFetchSize){
        this(sequenceName, format, batchFetchSize, prefix, prefixDelimiter, "", "");
    }

    /**
     * <p>
     * Constructs a generator that will return a String with a prefix
     * and the sequence value formatted according to <code>format</code>.
     * </p>
     * <p>
     * An example format is <code>"%05d"</code>. This will left pad the sequence
     * number up to 5 zeroes, 00001.
     * </p>
     * <p>
     * See <a href="http://docs.oracle.com/javase/7/docs/api/java/util/Formatter.html#syntax">
     * String.format</a> for further information on formatting options.
     * </p>
     * @param entityManager JPA entityManager used to create the query
     * @param sequenceName  the name of the sequence to get the numbers from
     * @param format how the sequence value is formatted as String
     * @param batchFetchSize how many values to fetch from database per call
     * @param postFix fixed string after sequence value
     * @param postFixDelimiter fixed string between postfix and sequence value
     */
    public PostgreSqlSequenceGenerator(String sequenceName, String postFix,
            String postFixDelimiter, long batchFetchSize, String format){
        this(sequenceName, format, batchFetchSize, "", "", postFix, postFixDelimiter);
    }

    @SuppressWarnings("unchecked")
    public final String next() {
        if (!iter.hasNext()) {
            Query query = entityManager.createNativeQuery("SELECT nextval( '" + sequenceName + "' ) FROM generate_series( 1, " + batchSize + " )");

            iter = (Iterator<BigInteger>) query.getResultList().iterator();
        }
        String value = String.format(format, iter.next().longValue());
        return prefix + prefixDelimiter + value + postFixDelimiter + postFix;
    }
}
